// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C89
// from https://raw.githubusercontent.com/FortAwesome/Font-Awesome/fa-4/src/icons.yml
// for use with https://github.com/FortAwesome/Font-Awesome/blob/fa-4/fonts/fontawesome-webfont.ttf
#pragma once

#define FONT_ICON_FILE_NAME_FA "fontawesome-webfont.ttf"

#define ICON_MIN_FA 0xf000
#define ICON_MAX_FA 0xf2e0
#define ICON_FA_GLASS "\xef\x80\x80"
#define ICON_FA_MUSIC "\xef\x80\x81"
#define ICON_FA_SEARCH "\xef\x80\x82"
#define ICON_FA_ENVELOPE_O "\xef\x80\x83"
#define ICON_FA_HEART "\xef\x80\x84"
#define ICON_FA_STAR "\xef\x80\x85"
#define ICON_FA_STAR_O "\xef\x80\x86"
#define ICON_FA_USER "\xef\x80\x87"
#define ICON_FA_FILM "\xef\x80\x88"
#define ICON_FA_TH_LARGE "\xef\x80\x89"
#define ICON_FA_TH "\xef\x80\x8a"
#define ICON_FA_TH_LIST "\xef\x80\x8b"
#define ICON_FA_CHECK "\xef\x80\x8c"
#define ICON_FA_TIMES "\xef\x80\x8d"
#define ICON_FA_SEARCH_PLUS "\xef\x80\x8e"
#define ICON_FA_SEARCH_MINUS "\xef\x80\x90"
#define ICON_FA_POWER_OFF "\xef\x80\x91"
#define ICON_FA_SIGNAL "\xef\x80\x92"
#define ICON_FA_COG "\xef\x80\x93"
#define ICON_FA_TRASH_O "\xef\x80\x94"
#define ICON_FA_HOME "\xef\x80\x95"
#define ICON_FA_FILE_O "\xef\x80\x96"
#define ICON_FA_CLOCK_O "\xef\x80\x97"
#define ICON_FA_ROAD "\xef\x80\x98"
#define ICON_FA_DOWNLOAD "\xef\x80\x99"
#define ICON_FA_ARROW_CIRCLE_O_DOWN "\xef\x80\x9a"
#define ICON_FA_ARROW_CIRCLE_O_UP "\xef\x80\x9b"
#define ICON_FA_INBOX "\xef\x80\x9c"
#define ICON_FA_PLAY_CIRCLE_O "\xef\x80\x9d"
#define ICON_FA_REPEAT "\xef\x80\x9e"
#define ICON_FA_REFRESH "\xef\x80\xa1"
#define ICON_FA_LIST_ALT "\xef\x80\xa2"
#define ICON_FA_LOCK "\xef\x80\xa3"
#define ICON_FA_FLAG "\xef\x80\xa4"
#define ICON_FA_HEADPHONES "\xef\x80\xa5"
#define ICON_FA_VOLUME_OFF "\xef\x80\xa6"
#define ICON_FA_VOLUME_DOWN "\xef\x80\xa7"
#define ICON_FA_VOLUME_UP "\xef\x80\xa8"
#define ICON_FA_QRCODE "\xef\x80\xa9"
#define ICON_FA_BARCODE "\xef\x80\xaa"
#define ICON_FA_TAG "\xef\x80\xab"
#define ICON_FA_TAGS "\xef\x80\xac"
#define ICON_FA_BOOK "\xef\x80\xad"
#define ICON_FA_BOOKMARK "\xef\x80\xae"
#define ICON_FA_PRINT "\xef\x80\xaf"
#define ICON_FA_CAMERA "\xef\x80\xb0"
#define ICON_FA_FONT "\xef\x80\xb1"
#define ICON_FA_BOLD "\xef\x80\xb2"
#define ICON_FA_ITALIC "\xef\x80\xb3"
#define ICON_FA_TEXT_HEIGHT "\xef\x80\xb4"
#define ICON_FA_TEXT_WIDTH "\xef\x80\xb5"
#define ICON_FA_ALIGN_LEFT "\xef\x80\xb6"
#define ICON_FA_ALIGN_CENTER "\xef\x80\xb7"
#define ICON_FA_ALIGN_RIGHT "\xef\x80\xb8"
#define ICON_FA_ALIGN_JUSTIFY "\xef\x80\xb9"
#define ICON_FA_LIST "\xef\x80\xba"
#define ICON_FA_OUTDENT "\xef\x80\xbb"
#define ICON_FA_INDENT "\xef\x80\xbc"
#define ICON_FA_VIDEO_CAMERA "\xef\x80\xbd"
#define ICON_FA_PICTURE_O "\xef\x80\xbe"
#define ICON_FA_PENCIL "\xef\x81\x80"
#define ICON_FA_MAP_MARKER "\xef\x81\x81"
#define ICON_FA_ADJUST "\xef\x81\x82"
#define ICON_FA_TINT "\xef\x81\x83"
#define ICON_FA_PENCIL_SQUARE_O "\xef\x81\x84"
#define ICON_FA_SHARE_SQUARE_O "\xef\x81\x85"
#define ICON_FA_CHECK_SQUARE_O "\xef\x81\x86"
#define ICON_FA_ARROWS "\xef\x81\x87"
#define ICON_FA_STEP_BACKWARD "\xef\x81\x88"
#define ICON_FA_FAST_BACKWARD "\xef\x81\x89"
#define ICON_FA_BACKWARD "\xef\x81\x8a"
#define ICON_FA_PLAY "\xef\x81\x8b"
#define ICON_FA_PAUSE "\xef\x81\x8c"
#define ICON_FA_STOP "\xef\x81\x8d"
#define ICON_FA_FORWARD "\xef\x81\x8e"
#define ICON_FA_FAST_FORWARD "\xef\x81\x90"
#define ICON_FA_STEP_FORWARD "\xef\x81\x91"
#define ICON_FA_EJECT "\xef\x81\x92"
#define ICON_FA_CHEVRON_LEFT "\xef\x81\x93"
#define ICON_FA_CHEVRON_RIGHT "\xef\x81\x94"
#define ICON_FA_PLUS_CIRCLE "\xef\x81\x95"
#define ICON_FA_MINUS_CIRCLE "\xef\x81\x96"
#define ICON_FA_TIMES_CIRCLE "\xef\x81\x97"
#define ICON_FA_CHECK_CIRCLE "\xef\x81\x98"
#define ICON_FA_QUESTION_CIRCLE "\xef\x81\x99"
#define ICON_FA_INFO_CIRCLE "\xef\x81\x9a"
#define ICON_FA_CROSSHAIRS "\xef\x81\x9b"
#define ICON_FA_TIMES_CIRCLE_O "\xef\x81\x9c"
#define ICON_FA_CHECK_CIRCLE_O "\xef\x81\x9d"
#define ICON_FA_BAN "\xef\x81\x9e"
#define ICON_FA_ARROW_LEFT "\xef\x81\xa0"
#define ICON_FA_ARROW_RIGHT "\xef\x81\xa1"
#define ICON_FA_ARROW_UP "\xef\x81\xa2"
#define ICON_FA_ARROW_DOWN "\xef\x81\xa3"
#define ICON_FA_SHARE "\xef\x81\xa4"
#define ICON_FA_EXPAND "\xef\x81\xa5"
#define ICON_FA_COMPRESS "\xef\x81\xa6"
#define ICON_FA_PLUS "\xef\x81\xa7"
#define ICON_FA_MINUS "\xef\x81\xa8"
#define ICON_FA_ASTERISK "\xef\x81\xa9"
#define ICON_FA_EXCLAMATION_CIRCLE "\xef\x81\xaa"
#define ICON_FA_GIFT "\xef\x81\xab"
#define ICON_FA_LEAF "\xef\x81\xac"
#define ICON_FA_FIRE "\xef\x81\xad"
#define ICON_FA_EYE "\xef\x81\xae"
#define ICON_FA_EYE_SLASH "\xef\x81\xb0"
#define ICON_FA_EXCLAMATION_TRIANGLE "\xef\x81\xb1"
#define ICON_FA_PLANE "\xef\x81\xb2"
#define ICON_FA_CALENDAR "\xef\x81\xb3"
#define ICON_FA_RANDOM "\xef\x81\xb4"
#define ICON_FA_COMMENT "\xef\x81\xb5"
#define ICON_FA_MAGNET "\xef\x81\xb6"
#define ICON_FA_CHEVRON_UP "\xef\x81\xb7"
#define ICON_FA_CHEVRON_DOWN "\xef\x81\xb8"
#define ICON_FA_RETWEET "\xef\x81\xb9"
#define ICON_FA_SHOPPING_CART "\xef\x81\xba"
#define ICON_FA_FOLDER "\xef\x81\xbb"
#define ICON_FA_FOLDER_OPEN "\xef\x81\xbc"
#define ICON_FA_ARROWS_V "\xef\x81\xbd"
#define ICON_FA_ARROWS_H "\xef\x81\xbe"
#define ICON_FA_BAR_CHART "\xef\x82\x80"
#define ICON_FA_TWITTER_SQUARE "\xef\x82\x81"
#define ICON_FA_FACEBOOK_SQUARE "\xef\x82\x82"
#define ICON_FA_CAMERA_RETRO "\xef\x82\x83"
#define ICON_FA_KEY "\xef\x82\x84"
#define ICON_FA_COGS "\xef\x82\x85"
#define ICON_FA_COMMENTS "\xef\x82\x86"
#define ICON_FA_THUMBS_O_UP "\xef\x82\x87"
#define ICON_FA_THUMBS_O_DOWN "\xef\x82\x88"
#define ICON_FA_STAR_HALF "\xef\x82\x89"
#define ICON_FA_HEART_O "\xef\x82\x8a"
#define ICON_FA_SIGN_OUT "\xef\x82\x8b"
#define ICON_FA_LINKEDIN_SQUARE "\xef\x82\x8c"
#define ICON_FA_THUMB_TACK "\xef\x82\x8d"
#define ICON_FA_EXTERNAL_LINK "\xef\x82\x8e"
#define ICON_FA_SIGN_IN "\xef\x82\x90"
#define ICON_FA_TROPHY "\xef\x82\x91"
#define ICON_FA_GITHUB_SQUARE "\xef\x82\x92"
#define ICON_FA_UPLOAD "\xef\x82\x93"
#define ICON_FA_LEMON_O "\xef\x82\x94"
#define ICON_FA_PHONE "\xef\x82\x95"
#define ICON_FA_SQUARE_O "\xef\x82\x96"
#define ICON_FA_BOOKMARK_O "\xef\x82\x97"
#define ICON_FA_PHONE_SQUARE "\xef\x82\x98"
#define ICON_FA_TWITTER "\xef\x82\x99"
#define ICON_FA_FACEBOOK "\xef\x82\x9a"
#define ICON_FA_GITHUB "\xef\x82\x9b"
#define ICON_FA_UNLOCK "\xef\x82\x9c"
#define ICON_FA_CREDIT_CARD "\xef\x82\x9d"
#define ICON_FA_RSS "\xef\x82\x9e"
#define ICON_FA_HDD_O "\xef\x82\xa0"
#define ICON_FA_BULLHORN "\xef\x82\xa1"
#define ICON_FA_BELL "\xef\x83\xb3"
#define ICON_FA_CERTIFICATE "\xef\x82\xa3"
#define ICON_FA_HAND_O_RIGHT "\xef\x82\xa4"
#define ICON_FA_HAND_O_LEFT "\xef\x82\xa5"
#define ICON_FA_HAND_O_UP "\xef\x82\xa6"
#define ICON_FA_HAND_O_DOWN "\xef\x82\xa7"
#define ICON_FA_ARROW_CIRCLE_LEFT "\xef\x82\xa8"
#define ICON_FA_ARROW_CIRCLE_RIGHT "\xef\x82\xa9"
#define ICON_FA_ARROW_CIRCLE_UP "\xef\x82\xaa"
#define ICON_FA_ARROW_CIRCLE_DOWN "\xef\x82\xab"
#define ICON_FA_GLOBE "\xef\x82\xac"
#define ICON_FA_WRENCH "\xef\x82\xad"
#define ICON_FA_TASKS "\xef\x82\xae"
#define ICON_FA_FILTER "\xef\x82\xb0"
#define ICON_FA_BRIEFCASE "\xef\x82\xb1"
#define ICON_FA_ARROWS_ALT "\xef\x82\xb2"
#define ICON_FA_USERS "\xef\x83\x80"
#define ICON_FA_LINK "\xef\x83\x81"
#define ICON_FA_CLOUD "\xef\x83\x82"
#define ICON_FA_FLASK "\xef\x83\x83"
#define ICON_FA_SCISSORS "\xef\x83\x84"
#define ICON_FA_FILES_O "\xef\x83\x85"
#define ICON_FA_PAPERCLIP "\xef\x83\x86"
#define ICON_FA_FLOPPY_O "\xef\x83\x87"
#define ICON_FA_SQUARE "\xef\x83\x88"
#define ICON_FA_BARS "\xef\x83\x89"
#define ICON_FA_LIST_UL "\xef\x83\x8a"
#define ICON_FA_LIST_OL "\xef\x83\x8b"
#define ICON_FA_STRIKETHROUGH "\xef\x83\x8c"
#define ICON_FA_UNDERLINE "\xef\x83\x8d"
#define ICON_FA_TABLE "\xef\x83\x8e"
#define ICON_FA_MAGIC "\xef\x83\x90"
#define ICON_FA_TRUCK "\xef\x83\x91"
#define ICON_FA_PINTEREST "\xef\x83\x92"
#define ICON_FA_PINTEREST_SQUARE "\xef\x83\x93"
#define ICON_FA_GOOGLE_PLUS_SQUARE "\xef\x83\x94"
#define ICON_FA_GOOGLE_PLUS "\xef\x83\x95"
#define ICON_FA_MONEY "\xef\x83\x96"
#define ICON_FA_CARET_DOWN "\xef\x83\x97"
#define ICON_FA_CARET_UP "\xef\x83\x98"
#define ICON_FA_CARET_LEFT "\xef\x83\x99"
#define ICON_FA_CARET_RIGHT "\xef\x83\x9a"
#define ICON_FA_COLUMNS "\xef\x83\x9b"
#define ICON_FA_SORT "\xef\x83\x9c"
#define ICON_FA_SORT_DESC "\xef\x83\x9d"
#define ICON_FA_SORT_ASC "\xef\x83\x9e"
#define ICON_FA_ENVELOPE "\xef\x83\xa0"
#define ICON_FA_LINKEDIN "\xef\x83\xa1"
#define ICON_FA_UNDO "\xef\x83\xa2"
#define ICON_FA_GAVEL "\xef\x83\xa3"
#define ICON_FA_TACHOMETER "\xef\x83\xa4"
#define ICON_FA_COMMENT_O "\xef\x83\xa5"
#define ICON_FA_COMMENTS_O "\xef\x83\xa6"
#define ICON_FA_BOLT "\xef\x83\xa7"
#define ICON_FA_SITEMAP "\xef\x83\xa8"
#define ICON_FA_UMBRELLA "\xef\x83\xa9"
#define ICON_FA_CLIPBOARD "\xef\x83\xaa"
#define ICON_FA_LIGHTBULB_O "\xef\x83\xab"
#define ICON_FA_EXCHANGE "\xef\x83\xac"
#define ICON_FA_CLOUD_DOWNLOAD "\xef\x83\xad"
#define ICON_FA_CLOUD_UPLOAD "\xef\x83\xae"
#define ICON_FA_USER_MD "\xef\x83\xb0"
#define ICON_FA_STETHOSCOPE "\xef\x83\xb1"
#define ICON_FA_SUITCASE "\xef\x83\xb2"
#define ICON_FA_BELL_O "\xef\x82\xa2"
#define ICON_FA_COFFEE "\xef\x83\xb4"
#define ICON_FA_CUTLERY "\xef\x83\xb5"
#define ICON_FA_FILE_TEXT_O "\xef\x83\xb6"
#define ICON_FA_BUILDING_O "\xef\x83\xb7"
#define ICON_FA_HOSPITAL_O "\xef\x83\xb8"
#define ICON_FA_AMBULANCE "\xef\x83\xb9"
#define ICON_FA_MEDKIT "\xef\x83\xba"
#define ICON_FA_FIGHTER_JET "\xef\x83\xbb"
#define ICON_FA_BEER "\xef\x83\xbc"
#define ICON_FA_H_SQUARE "\xef\x83\xbd"
#define ICON_FA_PLUS_SQUARE "\xef\x83\xbe"
#define ICON_FA_ANGLE_DOUBLE_LEFT "\xef\x84\x80"
#define ICON_FA_ANGLE_DOUBLE_RIGHT "\xef\x84\x81"
#define ICON_FA_ANGLE_DOUBLE_UP "\xef\x84\x82"
#define ICON_FA_ANGLE_DOUBLE_DOWN "\xef\x84\x83"
#define ICON_FA_ANGLE_LEFT "\xef\x84\x84"
#define ICON_FA_ANGLE_RIGHT "\xef\x84\x85"
#define ICON_FA_ANGLE_UP "\xef\x84\x86"
#define ICON_FA_ANGLE_DOWN "\xef\x84\x87"
#define ICON_FA_DESKTOP "\xef\x84\x88"
#define ICON_FA_LAPTOP "\xef\x84\x89"
#define ICON_FA_TABLET "\xef\x84\x8a"
#define ICON_FA_MOBILE "\xef\x84\x8b"
#define ICON_FA_CIRCLE_O "\xef\x84\x8c"
#define ICON_FA_QUOTE_LEFT "\xef\x84\x8d"
#define ICON_FA_QUOTE_RIGHT "\xef\x84\x8e"
#define ICON_FA_SPINNER "\xef\x84\x90"
#define ICON_FA_CIRCLE "\xef\x84\x91"
#define ICON_FA_REPLY "\xef\x84\x92"
#define ICON_FA_GITHUB_ALT "\xef\x84\x93"
#define ICON_FA_FOLDER_O "\xef\x84\x94"
#define ICON_FA_FOLDER_OPEN_O "\xef\x84\x95"
#define ICON_FA_SMILE_O "\xef\x84\x98"
#define ICON_FA_FROWN_O "\xef\x84\x99"
#define ICON_FA_MEH_O "\xef\x84\x9a"
#define ICON_FA_GAMEPAD "\xef\x84\x9b"
#define ICON_FA_KEYBOARD_O "\xef\x84\x9c"
#define ICON_FA_FLAG_O "\xef\x84\x9d"
#define ICON_FA_FLAG_CHECKERED "\xef\x84\x9e"
#define ICON_FA_TERMINAL "\xef\x84\xa0"
#define ICON_FA_CODE "\xef\x84\xa1"
#define ICON_FA_REPLY_ALL "\xef\x84\xa2"
#define ICON_FA_STAR_HALF_O "\xef\x84\xa3"
#define ICON_FA_LOCATION_ARROW "\xef\x84\xa4"
#define ICON_FA_CROP "\xef\x84\xa5"
#define ICON_FA_CODE_FORK "\xef\x84\xa6"
#define ICON_FA_CHAIN_BROKEN "\xef\x84\xa7"
#define ICON_FA_QUESTION "\xef\x84\xa8"
#define ICON_FA_INFO "\xef\x84\xa9"
#define ICON_FA_EXCLAMATION "\xef\x84\xaa"
#define ICON_FA_SUPERSCRIPT "\xef\x84\xab"
#define ICON_FA_SUBSCRIPT "\xef\x84\xac"
#define ICON_FA_ERASER "\xef\x84\xad"
#define ICON_FA_PUZZLE_PIECE "\xef\x84\xae"
#define ICON_FA_MICROPHONE "\xef\x84\xb0"
#define ICON_FA_MICROPHONE_SLASH "\xef\x84\xb1"
#define ICON_FA_SHIELD "\xef\x84\xb2"
#define ICON_FA_CALENDAR_O "\xef\x84\xb3"
#define ICON_FA_FIRE_EXTINGUISHER "\xef\x84\xb4"
#define ICON_FA_ROCKET "\xef\x84\xb5"
#define ICON_FA_MAXCDN "\xef\x84\xb6"
#define ICON_FA_CHEVRON_CIRCLE_LEFT "\xef\x84\xb7"
#define ICON_FA_CHEVRON_CIRCLE_RIGHT "\xef\x84\xb8"
#define ICON_FA_CHEVRON_CIRCLE_UP "\xef\x84\xb9"
#define ICON_FA_CHEVRON_CIRCLE_DOWN "\xef\x84\xba"
#define ICON_FA_HTML5 "\xef\x84\xbb"
#define ICON_FA_CSS3 "\xef\x84\xbc"
#define ICON_FA_ANCHOR "\xef\x84\xbd"
#define ICON_FA_UNLOCK_ALT "\xef\x84\xbe"
#define ICON_FA_BULLSEYE "\xef\x85\x80"
#define ICON_FA_ELLIPSIS_H "\xef\x85\x81"
#define ICON_FA_ELLIPSIS_V "\xef\x85\x82"
#define ICON_FA_RSS_SQUARE "\xef\x85\x83"
#define ICON_FA_PLAY_CIRCLE "\xef\x85\x84"
#define ICON_FA_TICKET "\xef\x85\x85"
#define ICON_FA_MINUS_SQUARE "\xef\x85\x86"
#define ICON_FA_MINUS_SQUARE_O "\xef\x85\x87"
#define ICON_FA_LEVEL_UP "\xef\x85\x88"
#define ICON_FA_LEVEL_DOWN "\xef\x85\x89"
#define ICON_FA_CHECK_SQUARE "\xef\x85\x8a"
#define ICON_FA_PENCIL_SQUARE "\xef\x85\x8b"
#define ICON_FA_EXTERNAL_LINK_SQUARE "\xef\x85\x8c"
#define ICON_FA_SHARE_SQUARE "\xef\x85\x8d"
#define ICON_FA_COMPASS "\xef\x85\x8e"
#define ICON_FA_CARET_SQUARE_O_DOWN "\xef\x85\x90"
#define ICON_FA_CARET_SQUARE_O_UP "\xef\x85\x91"
#define ICON_FA_CARET_SQUARE_O_RIGHT "\xef\x85\x92"
#define ICON_FA_EUR "\xef\x85\x93"
#define ICON_FA_GBP "\xef\x85\x94"
#define ICON_FA_USD "\xef\x85\x95"
#define ICON_FA_INR "\xef\x85\x96"
#define ICON_FA_JPY "\xef\x85\x97"
#define ICON_FA_RUB "\xef\x85\x98"
#define ICON_FA_KRW "\xef\x85\x99"
#define ICON_FA_BTC "\xef\x85\x9a"
#define ICON_FA_FILE "\xef\x85\x9b"
#define ICON_FA_FILE_TEXT "\xef\x85\x9c"
#define ICON_FA_SORT_ALPHA_ASC "\xef\x85\x9d"
#define ICON_FA_SORT_ALPHA_DESC "\xef\x85\x9e"
#define ICON_FA_SORT_AMOUNT_ASC "\xef\x85\xa0"
#define ICON_FA_SORT_AMOUNT_DESC "\xef\x85\xa1"
#define ICON_FA_SORT_NUMERIC_ASC "\xef\x85\xa2"
#define ICON_FA_SORT_NUMERIC_DESC "\xef\x85\xa3"
#define ICON_FA_THUMBS_UP "\xef\x85\xa4"
#define ICON_FA_THUMBS_DOWN "\xef\x85\xa5"
#define ICON_FA_YOUTUBE_SQUARE "\xef\x85\xa6"
#define ICON_FA_YOUTUBE "\xef\x85\xa7"
#define ICON_FA_XING "\xef\x85\xa8"
#define ICON_FA_XING_SQUARE "\xef\x85\xa9"
#define ICON_FA_YOUTUBE_PLAY "\xef\x85\xaa"
#define ICON_FA_DROPBOX "\xef\x85\xab"
#define ICON_FA_STACK_OVERFLOW "\xef\x85\xac"
#define ICON_FA_INSTAGRAM "\xef\x85\xad"
#define ICON_FA_FLICKR "\xef\x85\xae"
#define ICON_FA_ADN "\xef\x85\xb0"
#define ICON_FA_BITBUCKET "\xef\x85\xb1"
#define ICON_FA_BITBUCKET_SQUARE "\xef\x85\xb2"
#define ICON_FA_TUMBLR "\xef\x85\xb3"
#define ICON_FA_TUMBLR_SQUARE "\xef\x85\xb4"
#define ICON_FA_LONG_ARROW_DOWN "\xef\x85\xb5"
#define ICON_FA_LONG_ARROW_UP "\xef\x85\xb6"
#define ICON_FA_LONG_ARROW_LEFT "\xef\x85\xb7"
#define ICON_FA_LONG_ARROW_RIGHT "\xef\x85\xb8"
#define ICON_FA_APPLE "\xef\x85\xb9"
#define ICON_FA_WINDOWS "\xef\x85\xba"
#define ICON_FA_ANDROID "\xef\x85\xbb"
#define ICON_FA_LINUX "\xef\x85\xbc"
#define ICON_FA_DRIBBBLE "\xef\x85\xbd"
#define ICON_FA_SKYPE "\xef\x85\xbe"
#define ICON_FA_FOURSQUARE "\xef\x86\x80"
#define ICON_FA_TRELLO "\xef\x86\x81"
#define ICON_FA_FEMALE "\xef\x86\x82"
#define ICON_FA_MALE "\xef\x86\x83"
#define ICON_FA_GRATIPAY "\xef\x86\x84"
#define ICON_FA_SUN_O "\xef\x86\x85"
#define ICON_FA_MOON_O "\xef\x86\x86"
#define ICON_FA_ARCHIVE "\xef\x86\x87"
#define ICON_FA_BUG "\xef\x86\x88"
#define ICON_FA_VK "\xef\x86\x89"
#define ICON_FA_WEIBO "\xef\x86\x8a"
#define ICON_FA_RENREN "\xef\x86\x8b"
#define ICON_FA_PAGELINES "\xef\x86\x8c"
#define ICON_FA_STACK_EXCHANGE "\xef\x86\x8d"
#define ICON_FA_ARROW_CIRCLE_O_RIGHT "\xef\x86\x8e"
#define ICON_FA_ARROW_CIRCLE_O_LEFT "\xef\x86\x90"
#define ICON_FA_CARET_SQUARE_O_LEFT "\xef\x86\x91"
#define ICON_FA_DOT_CIRCLE_O "\xef\x86\x92"
#define ICON_FA_WHEELCHAIR "\xef\x86\x93"
#define ICON_FA_VIMEO_SQUARE "\xef\x86\x94"
#define ICON_FA_TRY "\xef\x86\x95"
#define ICON_FA_PLUS_SQUARE_O "\xef\x86\x96"
#define ICON_FA_SPACE_SHUTTLE "\xef\x86\x97"
#define ICON_FA_SLACK "\xef\x86\x98"
#define ICON_FA_ENVELOPE_SQUARE "\xef\x86\x99"
#define ICON_FA_WORDPRESS "\xef\x86\x9a"
#define ICON_FA_OPENID "\xef\x86\x9b"
#define ICON_FA_UNIVERSITY "\xef\x86\x9c"
#define ICON_FA_GRADUATION_CAP "\xef\x86\x9d"
#define ICON_FA_YAHOO "\xef\x86\x9e"
#define ICON_FA_GOOGLE "\xef\x86\xa0"
#define ICON_FA_REDDIT "\xef\x86\xa1"
#define ICON_FA_REDDIT_SQUARE "\xef\x86\xa2"
#define ICON_FA_STUMBLEUPON_CIRCLE "\xef\x86\xa3"
#define ICON_FA_STUMBLEUPON "\xef\x86\xa4"
#define ICON_FA_DELICIOUS "\xef\x86\xa5"
#define ICON_FA_DIGG "\xef\x86\xa6"
#define ICON_FA_PIED_PIPER_PP "\xef\x86\xa7"
#define ICON_FA_PIED_PIPER_ALT "\xef\x86\xa8"
#define ICON_FA_DRUPAL "\xef\x86\xa9"
#define ICON_FA_JOOMLA "\xef\x86\xaa"
#define ICON_FA_LANGUAGE "\xef\x86\xab"
#define ICON_FA_FAX "\xef\x86\xac"
#define ICON_FA_BUILDING "\xef\x86\xad"
#define ICON_FA_CHILD "\xef\x86\xae"
#define ICON_FA_PAW "\xef\x86\xb0"
#define ICON_FA_SPOON "\xef\x86\xb1"
#define ICON_FA_CUBE "\xef\x86\xb2"
#define ICON_FA_CUBES "\xef\x86\xb3"
#define ICON_FA_BEHANCE "\xef\x86\xb4"
#define ICON_FA_BEHANCE_SQUARE "\xef\x86\xb5"
#define ICON_FA_STEAM "\xef\x86\xb6"
#define ICON_FA_STEAM_SQUARE "\xef\x86\xb7"
#define ICON_FA_RECYCLE "\xef\x86\xb8"
#define ICON_FA_CAR "\xef\x86\xb9"
#define ICON_FA_TAXI "\xef\x86\xba"
#define ICON_FA_TREE "\xef\x86\xbb"
#define ICON_FA_SPOTIFY "\xef\x86\xbc"
#define ICON_FA_DEVIANTART "\xef\x86\xbd"
#define ICON_FA_SOUNDCLOUD "\xef\x86\xbe"
#define ICON_FA_DATABASE "\xef\x87\x80"
#define ICON_FA_FILE_PDF_O "\xef\x87\x81"
#define ICON_FA_FILE_WORD_O "\xef\x87\x82"
#define ICON_FA_FILE_EXCEL_O "\xef\x87\x83"
#define ICON_FA_FILE_POWERPOINT_O "\xef\x87\x84"
#define ICON_FA_FILE_IMAGE_O "\xef\x87\x85"
#define ICON_FA_FILE_ARCHIVE_O "\xef\x87\x86"
#define ICON_FA_FILE_AUDIO_O "\xef\x87\x87"
#define ICON_FA_FILE_VIDEO_O "\xef\x87\x88"
#define ICON_FA_FILE_CODE_O "\xef\x87\x89"
#define ICON_FA_VINE "\xef\x87\x8a"
#define ICON_FA_CODEPEN "\xef\x87\x8b"
#define ICON_FA_JSFIDDLE "\xef\x87\x8c"
#define ICON_FA_LIFE_RING "\xef\x87\x8d"
#define ICON_FA_CIRCLE_O_NOTCH "\xef\x87\x8e"
#define ICON_FA_REBEL "\xef\x87\x90"
#define ICON_FA_EMPIRE "\xef\x87\x91"
#define ICON_FA_GIT_SQUARE "\xef\x87\x92"
#define ICON_FA_GIT "\xef\x87\x93"
#define ICON_FA_HACKER_NEWS "\xef\x87\x94"
#define ICON_FA_TENCENT_WEIBO "\xef\x87\x95"
#define ICON_FA_QQ "\xef\x87\x96"
#define ICON_FA_WEIXIN "\xef\x87\x97"
#define ICON_FA_PAPER_PLANE "\xef\x87\x98"
#define ICON_FA_PAPER_PLANE_O "\xef\x87\x99"
#define ICON_FA_HISTORY "\xef\x87\x9a"
#define ICON_FA_CIRCLE_THIN "\xef\x87\x9b"
#define ICON_FA_HEADER "\xef\x87\x9c"
#define ICON_FA_PARAGRAPH "\xef\x87\x9d"
#define ICON_FA_SLIDERS "\xef\x87\x9e"
#define ICON_FA_SHARE_ALT "\xef\x87\xa0"
#define ICON_FA_SHARE_ALT_SQUARE "\xef\x87\xa1"
#define ICON_FA_BOMB "\xef\x87\xa2"
#define ICON_FA_FUTBOL_O "\xef\x87\xa3"
#define ICON_FA_TTY "\xef\x87\xa4"
#define ICON_FA_BINOCULARS "\xef\x87\xa5"
#define ICON_FA_PLUG "\xef\x87\xa6"
#define ICON_FA_SLIDESHARE "\xef\x87\xa7"
#define ICON_FA_TWITCH "\xef\x87\xa8"
#define ICON_FA_YELP "\xef\x87\xa9"
#define ICON_FA_NEWSPAPER_O "\xef\x87\xaa"
#define ICON_FA_WIFI "\xef\x87\xab"
#define ICON_FA_CALCULATOR "\xef\x87\xac"
#define ICON_FA_PAYPAL "\xef\x87\xad"
#define ICON_FA_GOOGLE_WALLET "\xef\x87\xae"
#define ICON_FA_CC_VISA "\xef\x87\xb0"
#define ICON_FA_CC_MASTERCARD "\xef\x87\xb1"
#define ICON_FA_CC_DISCOVER "\xef\x87\xb2"
#define ICON_FA_CC_AMEX "\xef\x87\xb3"
#define ICON_FA_CC_PAYPAL "\xef\x87\xb4"
#define ICON_FA_CC_STRIPE "\xef\x87\xb5"
#define ICON_FA_BELL_SLASH "\xef\x87\xb6"
#define ICON_FA_BELL_SLASH_O "\xef\x87\xb7"
#define ICON_FA_TRASH "\xef\x87\xb8"
#define ICON_FA_COPYRIGHT "\xef\x87\xb9"
#define ICON_FA_AT "\xef\x87\xba"
#define ICON_FA_EYEDROPPER "\xef\x87\xbb"
#define ICON_FA_PAINT_BRUSH "\xef\x87\xbc"
#define ICON_FA_BIRTHDAY_CAKE "\xef\x87\xbd"
#define ICON_FA_AREA_CHART "\xef\x87\xbe"
#define ICON_FA_PIE_CHART "\xef\x88\x80"
#define ICON_FA_LINE_CHART "\xef\x88\x81"
#define ICON_FA_LASTFM "\xef\x88\x82"
#define ICON_FA_LASTFM_SQUARE "\xef\x88\x83"
#define ICON_FA_TOGGLE_OFF "\xef\x88\x84"
#define ICON_FA_TOGGLE_ON "\xef\x88\x85"
#define ICON_FA_BICYCLE "\xef\x88\x86"
#define ICON_FA_BUS "\xef\x88\x87"
#define ICON_FA_IOXHOST "\xef\x88\x88"
#define ICON_FA_ANGELLIST "\xef\x88\x89"
#define ICON_FA_CC "\xef\x88\x8a"
#define ICON_FA_ILS "\xef\x88\x8b"
#define ICON_FA_MEANPATH "\xef\x88\x8c"
#define ICON_FA_BUYSELLADS "\xef\x88\x8d"
#define ICON_FA_CONNECTDEVELOP "\xef\x88\x8e"
#define ICON_FA_DASHCUBE "\xef\x88\x90"
#define ICON_FA_FORUMBEE "\xef\x88\x91"
#define ICON_FA_LEANPUB "\xef\x88\x92"
#define ICON_FA_SELLSY "\xef\x88\x93"
#define ICON_FA_SHIRTSINBULK "\xef\x88\x94"
#define ICON_FA_SIMPLYBUILT "\xef\x88\x95"
#define ICON_FA_SKYATLAS "\xef\x88\x96"
#define ICON_FA_CART_PLUS "\xef\x88\x97"
#define ICON_FA_CART_ARROW_DOWN "\xef\x88\x98"
#define ICON_FA_DIAMOND "\xef\x88\x99"
#define ICON_FA_SHIP "\xef\x88\x9a"
#define ICON_FA_USER_SECRET "\xef\x88\x9b"
#define ICON_FA_MOTORCYCLE "\xef\x88\x9c"
#define ICON_FA_STREET_VIEW "\xef\x88\x9d"
#define ICON_FA_HEARTBEAT "\xef\x88\x9e"
#define ICON_FA_VENUS "\xef\x88\xa1"
#define ICON_FA_MARS "\xef\x88\xa2"
#define ICON_FA_MERCURY "\xef\x88\xa3"
#define ICON_FA_TRANSGENDER "\xef\x88\xa4"
#define ICON_FA_TRANSGENDER_ALT "\xef\x88\xa5"
#define ICON_FA_VENUS_DOUBLE "\xef\x88\xa6"
#define ICON_FA_MARS_DOUBLE "\xef\x88\xa7"
#define ICON_FA_VENUS_MARS "\xef\x88\xa8"
#define ICON_FA_MARS_STROKE "\xef\x88\xa9"
#define ICON_FA_MARS_STROKE_V "\xef\x88\xaa"
#define ICON_FA_MARS_STROKE_H "\xef\x88\xab"
#define ICON_FA_NEUTER "\xef\x88\xac"
#define ICON_FA_GENDERLESS "\xef\x88\xad"
#define ICON_FA_FACEBOOK_OFFICIAL "\xef\x88\xb0"
#define ICON_FA_PINTEREST_P "\xef\x88\xb1"
#define ICON_FA_WHATSAPP "\xef\x88\xb2"
#define ICON_FA_SERVER "\xef\x88\xb3"
#define ICON_FA_USER_PLUS "\xef\x88\xb4"
#define ICON_FA_USER_TIMES "\xef\x88\xb5"
#define ICON_FA_BED "\xef\x88\xb6"
#define ICON_FA_VIACOIN "\xef\x88\xb7"
#define ICON_FA_TRAIN "\xef\x88\xb8"
#define ICON_FA_SUBWAY "\xef\x88\xb9"
#define ICON_FA_MEDIUM "\xef\x88\xba"
#define ICON_FA_Y_COMBINATOR "\xef\x88\xbb"
#define ICON_FA_OPTIN_MONSTER "\xef\x88\xbc"
#define ICON_FA_OPENCART "\xef\x88\xbd"
#define ICON_FA_EXPEDITEDSSL "\xef\x88\xbe"
#define ICON_FA_BATTERY_FULL "\xef\x89\x80"
#define ICON_FA_BATTERY_THREE_QUARTERS "\xef\x89\x81"
#define ICON_FA_BATTERY_HALF "\xef\x89\x82"
#define ICON_FA_BATTERY_QUARTER "\xef\x89\x83"
#define ICON_FA_BATTERY_EMPTY "\xef\x89\x84"
#define ICON_FA_MOUSE_POINTER "\xef\x89\x85"
#define ICON_FA_I_CURSOR "\xef\x89\x86"
#define ICON_FA_OBJECT_GROUP "\xef\x89\x87"
#define ICON_FA_OBJECT_UNGROUP "\xef\x89\x88"
#define ICON_FA_STICKY_NOTE "\xef\x89\x89"
#define ICON_FA_STICKY_NOTE_O "\xef\x89\x8a"
#define ICON_FA_CC_JCB "\xef\x89\x8b"
#define ICON_FA_CC_DINERS_CLUB "\xef\x89\x8c"
#define ICON_FA_CLONE "\xef\x89\x8d"
#define ICON_FA_BALANCE_SCALE "\xef\x89\x8e"
#define ICON_FA_HOURGLASS_O "\xef\x89\x90"
#define ICON_FA_HOURGLASS_START "\xef\x89\x91"
#define ICON_FA_HOURGLASS_HALF "\xef\x89\x92"
#define ICON_FA_HOURGLASS_END "\xef\x89\x93"
#define ICON_FA_HOURGLASS "\xef\x89\x94"
#define ICON_FA_HAND_ROCK_O "\xef\x89\x95"
#define ICON_FA_HAND_PAPER_O "\xef\x89\x96"
#define ICON_FA_HAND_SCISSORS_O "\xef\x89\x97"
#define ICON_FA_HAND_LIZARD_O "\xef\x89\x98"
#define ICON_FA_HAND_SPOCK_O "\xef\x89\x99"
#define ICON_FA_HAND_POINTER_O "\xef\x89\x9a"
#define ICON_FA_HAND_PEACE_O "\xef\x89\x9b"
#define ICON_FA_TRADEMARK "\xef\x89\x9c"
#define ICON_FA_REGISTERED "\xef\x89\x9d"
#define ICON_FA_CREATIVE_COMMONS "\xef\x89\x9e"
#define ICON_FA_GG "\xef\x89\xa0"
#define ICON_FA_GG_CIRCLE "\xef\x89\xa1"
#define ICON_FA_TRIPADVISOR "\xef\x89\xa2"
#define ICON_FA_ODNOKLASSNIKI "\xef\x89\xa3"
#define ICON_FA_ODNOKLASSNIKI_SQUARE "\xef\x89\xa4"
#define ICON_FA_GET_POCKET "\xef\x89\xa5"
#define ICON_FA_WIKIPEDIA_W "\xef\x89\xa6"
#define ICON_FA_SAFARI "\xef\x89\xa7"
#define ICON_FA_CHROME "\xef\x89\xa8"
#define ICON_FA_FIREFOX "\xef\x89\xa9"
#define ICON_FA_OPERA "\xef\x89\xaa"
#define ICON_FA_INTERNET_EXPLORER "\xef\x89\xab"
#define ICON_FA_TELEVISION "\xef\x89\xac"
#define ICON_FA_CONTAO "\xef\x89\xad"
#define ICON_FA_500PX "\xef\x89\xae"
#define ICON_FA_AMAZON "\xef\x89\xb0"
#define ICON_FA_CALENDAR_PLUS_O "\xef\x89\xb1"
#define ICON_FA_CALENDAR_MINUS_O "\xef\x89\xb2"
#define ICON_FA_CALENDAR_TIMES_O "\xef\x89\xb3"
#define ICON_FA_CALENDAR_CHECK_O "\xef\x89\xb4"
#define ICON_FA_INDUSTRY "\xef\x89\xb5"
#define ICON_FA_MAP_PIN "\xef\x89\xb6"
#define ICON_FA_MAP_SIGNS "\xef\x89\xb7"
#define ICON_FA_MAP_O "\xef\x89\xb8"
#define ICON_FA_MAP "\xef\x89\xb9"
#define ICON_FA_COMMENTING "\xef\x89\xba"
#define ICON_FA_COMMENTING_O "\xef\x89\xbb"
#define ICON_FA_HOUZZ "\xef\x89\xbc"
#define ICON_FA_VIMEO "\xef\x89\xbd"
#define ICON_FA_BLACK_TIE "\xef\x89\xbe"
#define ICON_FA_FONTICONS "\xef\x8a\x80"
#define ICON_FA_REDDIT_ALIEN "\xef\x8a\x81"
#define ICON_FA_EDGE "\xef\x8a\x82"
#define ICON_FA_CREDIT_CARD_ALT "\xef\x8a\x83"
#define ICON_FA_CODIEPIE "\xef\x8a\x84"
#define ICON_FA_MODX "\xef\x8a\x85"
#define ICON_FA_FORT_AWESOME "\xef\x8a\x86"
#define ICON_FA_USB "\xef\x8a\x87"
#define ICON_FA_PRODUCT_HUNT "\xef\x8a\x88"
#define ICON_FA_MIXCLOUD "\xef\x8a\x89"
#define ICON_FA_SCRIBD "\xef\x8a\x8a"
#define ICON_FA_PAUSE_CIRCLE "\xef\x8a\x8b"
#define ICON_FA_PAUSE_CIRCLE_O "\xef\x8a\x8c"
#define ICON_FA_STOP_CIRCLE "\xef\x8a\x8d"
#define ICON_FA_STOP_CIRCLE_O "\xef\x8a\x8e"
#define ICON_FA_SHOPPING_BAG "\xef\x8a\x90"
#define ICON_FA_SHOPPING_BASKET "\xef\x8a\x91"
#define ICON_FA_HASHTAG "\xef\x8a\x92"
#define ICON_FA_BLUETOOTH "\xef\x8a\x93"
#define ICON_FA_BLUETOOTH_B "\xef\x8a\x94"
#define ICON_FA_PERCENT "\xef\x8a\x95"
#define ICON_FA_GITLAB "\xef\x8a\x96"
#define ICON_FA_WPBEGINNER "\xef\x8a\x97"
#define ICON_FA_WPFORMS "\xef\x8a\x98"
#define ICON_FA_ENVIRA "\xef\x8a\x99"
#define ICON_FA_UNIVERSAL_ACCESS "\xef\x8a\x9a"
#define ICON_FA_WHEELCHAIR_ALT "\xef\x8a\x9b"
#define ICON_FA_QUESTION_CIRCLE_O "\xef\x8a\x9c"
#define ICON_FA_BLIND "\xef\x8a\x9d"
#define ICON_FA_AUDIO_DESCRIPTION "\xef\x8a\x9e"
#define ICON_FA_VOLUME_CONTROL_PHONE "\xef\x8a\xa0"
#define ICON_FA_BRAILLE "\xef\x8a\xa1"
#define ICON_FA_ASSISTIVE_LISTENING_SYSTEMS "\xef\x8a\xa2"
#define ICON_FA_AMERICAN_SIGN_LANGUAGE_INTERPRETING "\xef\x8a\xa3"
#define ICON_FA_DEAF "\xef\x8a\xa4"
#define ICON_FA_GLIDE "\xef\x8a\xa5"
#define ICON_FA_GLIDE_G "\xef\x8a\xa6"
#define ICON_FA_SIGN_LANGUAGE "\xef\x8a\xa7"
#define ICON_FA_LOW_VISION "\xef\x8a\xa8"
#define ICON_FA_VIADEO "\xef\x8a\xa9"
#define ICON_FA_VIADEO_SQUARE "\xef\x8a\xaa"
#define ICON_FA_SNAPCHAT "\xef\x8a\xab"
#define ICON_FA_SNAPCHAT_GHOST "\xef\x8a\xac"
#define ICON_FA_SNAPCHAT_SQUARE "\xef\x8a\xad"
#define ICON_FA_PIED_PIPER "\xef\x8a\xae"
#define ICON_FA_FIRST_ORDER "\xef\x8a\xb0"
#define ICON_FA_YOAST "\xef\x8a\xb1"
#define ICON_FA_THEMEISLE "\xef\x8a\xb2"
#define ICON_FA_GOOGLE_PLUS_OFFICIAL "\xef\x8a\xb3"
#define ICON_FA_FONT_AWESOME "\xef\x8a\xb4"
#define ICON_FA_HANDSHAKE_O "\xef\x8a\xb5"
#define ICON_FA_ENVELOPE_OPEN "\xef\x8a\xb6"
#define ICON_FA_ENVELOPE_OPEN_O "\xef\x8a\xb7"
#define ICON_FA_LINODE "\xef\x8a\xb8"
#define ICON_FA_ADDRESS_BOOK "\xef\x8a\xb9"
#define ICON_FA_ADDRESS_BOOK_O "\xef\x8a\xba"
#define ICON_FA_ADDRESS_CARD "\xef\x8a\xbb"
#define ICON_FA_ADDRESS_CARD_O "\xef\x8a\xbc"
#define ICON_FA_USER_CIRCLE "\xef\x8a\xbd"
#define ICON_FA_USER_CIRCLE_O "\xef\x8a\xbe"
#define ICON_FA_USER_O "\xef\x8b\x80"
#define ICON_FA_ID_BADGE "\xef\x8b\x81"
#define ICON_FA_ID_CARD "\xef\x8b\x82"
#define ICON_FA_ID_CARD_O "\xef\x8b\x83"
#define ICON_FA_QUORA "\xef\x8b\x84"
#define ICON_FA_FREE_CODE_CAMP "\xef\x8b\x85"
#define ICON_FA_TELEGRAM "\xef\x8b\x86"
#define ICON_FA_THERMOMETER_FULL "\xef\x8b\x87"
#define ICON_FA_THERMOMETER_THREE_QUARTERS "\xef\x8b\x88"
#define ICON_FA_THERMOMETER_HALF "\xef\x8b\x89"
#define ICON_FA_THERMOMETER_QUARTER "\xef\x8b\x8a"
#define ICON_FA_THERMOMETER_EMPTY "\xef\x8b\x8b"
#define ICON_FA_SHOWER "\xef\x8b\x8c"
#define ICON_FA_BATH "\xef\x8b\x8d"
#define ICON_FA_PODCAST "\xef\x8b\x8e"
#define ICON_FA_WINDOW_MAXIMIZE "\xef\x8b\x90"
#define ICON_FA_WINDOW_MINIMIZE "\xef\x8b\x91"
#define ICON_FA_WINDOW_RESTORE "\xef\x8b\x92"
#define ICON_FA_WINDOW_CLOSE "\xef\x8b\x93"
#define ICON_FA_WINDOW_CLOSE_O "\xef\x8b\x94"
#define ICON_FA_BANDCAMP "\xef\x8b\x95"
#define ICON_FA_GRAV "\xef\x8b\x96"
#define ICON_FA_ETSY "\xef\x8b\x97"
#define ICON_FA_IMDB "\xef\x8b\x98"
#define ICON_FA_RAVELRY "\xef\x8b\x99"
#define ICON_FA_EERCAST "\xef\x8b\x9a"
#define ICON_FA_MICROCHIP "\xef\x8b\x9b"
#define ICON_FA_SNOWFLAKE_O "\xef\x8b\x9c"
#define ICON_FA_SUPERPOWERS "\xef\x8b\x9d"
#define ICON_FA_WPEXPLORER "\xef\x8b\x9e"
#define ICON_FA_MEETUP "\xef\x8b\xa0"
