// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C89
// from https://raw.githubusercontent.com/google/material-design-icons/master/iconfont/codepoints
// for use with https://github.com/google/material-design-icons/blob/master/iconfont/MaterialIcons-Regular.ttf
#pragma once

#define FONT_ICON_FILE_NAME_MD "MaterialIcons-Regular.ttf"

#define ICON_MIN_MD 0xe000
#define ICON_MAX_MD 0xeb4c
#define ICON_MD_3D_ROTATION "\xee\xa1\x8d"
#define ICON_MD_AC_UNIT "\xee\xac\xbb"
#define ICON_MD_ACCESS_ALARM "\xee\x86\x90"
#define ICON_MD_ACCESS_ALARMS "\xee\x86\x91"
#define ICON_MD_ACCESS_TIME "\xee\x86\x92"
#define ICON_MD_ACCESSIBILITY "\xee\xa1\x8e"
#define ICON_MD_ACCESSIBLE "\xee\xa4\x94"
#define ICON_MD_ACCOUNT_BALANCE "\xee\xa1\x8f"
#define ICON_MD_ACCOUNT_BALANCE_WALLET "\xee\xa1\x90"
#define ICON_MD_ACCOUNT_BOX "\xee\xa1\x91"
#define ICON_MD_ACCOUNT_CIRCLE "\xee\xa1\x93"
#define ICON_MD_ADB "\xee\x98\x8e"
#define ICON_MD_ADD "\xee\x85\x85"
#define ICON_MD_ADD_A_PHOTO "\xee\x90\xb9"
#define ICON_MD_ADD_ALARM "\xee\x86\x93"
#define ICON_MD_ADD_ALERT "\xee\x80\x83"
#define ICON_MD_ADD_BOX "\xee\x85\x86"
#define ICON_MD_ADD_CIRCLE "\xee\x85\x87"
#define ICON_MD_ADD_CIRCLE_OUTLINE "\xee\x85\x88"
#define ICON_MD_ADD_LOCATION "\xee\x95\xa7"
#define ICON_MD_ADD_SHOPPING_CART "\xee\xa1\x94"
#define ICON_MD_ADD_TO_PHOTOS "\xee\x8e\x9d"
#define ICON_MD_ADD_TO_QUEUE "\xee\x81\x9c"
#define ICON_MD_ADJUST "\xee\x8e\x9e"
#define ICON_MD_AIRLINE_SEAT_FLAT "\xee\x98\xb0"
#define ICON_MD_AIRLINE_SEAT_FLAT_ANGLED "\xee\x98\xb1"
#define ICON_MD_AIRLINE_SEAT_INDIVIDUAL_SUITE "\xee\x98\xb2"
#define ICON_MD_AIRLINE_SEAT_LEGROOM_EXTRA "\xee\x98\xb3"
#define ICON_MD_AIRLINE_SEAT_LEGROOM_NORMAL "\xee\x98\xb4"
#define ICON_MD_AIRLINE_SEAT_LEGROOM_REDUCED "\xee\x98\xb5"
#define ICON_MD_AIRLINE_SEAT_RECLINE_EXTRA "\xee\x98\xb6"
#define ICON_MD_AIRLINE_SEAT_RECLINE_NORMAL "\xee\x98\xb7"
#define ICON_MD_AIRPLANEMODE_ACTIVE "\xee\x86\x95"
#define ICON_MD_AIRPLANEMODE_INACTIVE "\xee\x86\x94"
#define ICON_MD_AIRPLAY "\xee\x81\x95"
#define ICON_MD_AIRPORT_SHUTTLE "\xee\xac\xbc"
#define ICON_MD_ALARM "\xee\xa1\x95"
#define ICON_MD_ALARM_ADD "\xee\xa1\x96"
#define ICON_MD_ALARM_OFF "\xee\xa1\x97"
#define ICON_MD_ALARM_ON "\xee\xa1\x98"
#define ICON_MD_ALBUM "\xee\x80\x99"
#define ICON_MD_ALL_INCLUSIVE "\xee\xac\xbd"
#define ICON_MD_ALL_OUT "\xee\xa4\x8b"
#define ICON_MD_ANDROID "\xee\xa1\x99"
#define ICON_MD_ANNOUNCEMENT "\xee\xa1\x9a"
#define ICON_MD_APPS "\xee\x97\x83"
#define ICON_MD_ARCHIVE "\xee\x85\x89"
#define ICON_MD_ARROW_BACK "\xee\x97\x84"
#define ICON_MD_ARROW_DOWNWARD "\xee\x97\x9b"
#define ICON_MD_ARROW_DROP_DOWN "\xee\x97\x85"
#define ICON_MD_ARROW_DROP_DOWN_CIRCLE "\xee\x97\x86"
#define ICON_MD_ARROW_DROP_UP "\xee\x97\x87"
#define ICON_MD_ARROW_FORWARD "\xee\x97\x88"
#define ICON_MD_ARROW_UPWARD "\xee\x97\x98"
#define ICON_MD_ART_TRACK "\xee\x81\xa0"
#define ICON_MD_ASPECT_RATIO "\xee\xa1\x9b"
#define ICON_MD_ASSESSMENT "\xee\xa1\x9c"
#define ICON_MD_ASSIGNMENT "\xee\xa1\x9d"
#define ICON_MD_ASSIGNMENT_IND "\xee\xa1\x9e"
#define ICON_MD_ASSIGNMENT_LATE "\xee\xa1\x9f"
#define ICON_MD_ASSIGNMENT_RETURN "\xee\xa1\xa0"
#define ICON_MD_ASSIGNMENT_RETURNED "\xee\xa1\xa1"
#define ICON_MD_ASSIGNMENT_TURNED_IN "\xee\xa1\xa2"
#define ICON_MD_ASSISTANT "\xee\x8e\x9f"
#define ICON_MD_ASSISTANT_PHOTO "\xee\x8e\xa0"
#define ICON_MD_ATTACH_FILE "\xee\x88\xa6"
#define ICON_MD_ATTACH_MONEY "\xee\x88\xa7"
#define ICON_MD_ATTACHMENT "\xee\x8a\xbc"
#define ICON_MD_AUDIOTRACK "\xee\x8e\xa1"
#define ICON_MD_AUTORENEW "\xee\xa1\xa3"
#define ICON_MD_AV_TIMER "\xee\x80\x9b"
#define ICON_MD_BACKSPACE "\xee\x85\x8a"
#define ICON_MD_BACKUP "\xee\xa1\xa4"
#define ICON_MD_BATTERY_ALERT "\xee\x86\x9c"
#define ICON_MD_BATTERY_CHARGING_FULL "\xee\x86\xa3"
#define ICON_MD_BATTERY_FULL "\xee\x86\xa4"
#define ICON_MD_BATTERY_STD "\xee\x86\xa5"
#define ICON_MD_BATTERY_UNKNOWN "\xee\x86\xa6"
#define ICON_MD_BEACH_ACCESS "\xee\xac\xbe"
#define ICON_MD_BEENHERE "\xee\x94\xad"
#define ICON_MD_BLOCK "\xee\x85\x8b"
#define ICON_MD_BLUETOOTH "\xee\x86\xa7"
#define ICON_MD_BLUETOOTH_AUDIO "\xee\x98\x8f"
#define ICON_MD_BLUETOOTH_CONNECTED "\xee\x86\xa8"
#define ICON_MD_BLUETOOTH_DISABLED "\xee\x86\xa9"
#define ICON_MD_BLUETOOTH_SEARCHING "\xee\x86\xaa"
#define ICON_MD_BLUR_CIRCULAR "\xee\x8e\xa2"
#define ICON_MD_BLUR_LINEAR "\xee\x8e\xa3"
#define ICON_MD_BLUR_OFF "\xee\x8e\xa4"
#define ICON_MD_BLUR_ON "\xee\x8e\xa5"
#define ICON_MD_BOOK "\xee\xa1\xa5"
#define ICON_MD_BOOKMARK "\xee\xa1\xa6"
#define ICON_MD_BOOKMARK_BORDER "\xee\xa1\xa7"
#define ICON_MD_BORDER_ALL "\xee\x88\xa8"
#define ICON_MD_BORDER_BOTTOM "\xee\x88\xa9"
#define ICON_MD_BORDER_CLEAR "\xee\x88\xaa"
#define ICON_MD_BORDER_COLOR "\xee\x88\xab"
#define ICON_MD_BORDER_HORIZONTAL "\xee\x88\xac"
#define ICON_MD_BORDER_INNER "\xee\x88\xad"
#define ICON_MD_BORDER_LEFT "\xee\x88\xae"
#define ICON_MD_BORDER_OUTER "\xee\x88\xaf"
#define ICON_MD_BORDER_RIGHT "\xee\x88\xb0"
#define ICON_MD_BORDER_STYLE "\xee\x88\xb1"
#define ICON_MD_BORDER_TOP "\xee\x88\xb2"
#define ICON_MD_BORDER_VERTICAL "\xee\x88\xb3"
#define ICON_MD_BRANDING_WATERMARK "\xee\x81\xab"
#define ICON_MD_BRIGHTNESS_1 "\xee\x8e\xa6"
#define ICON_MD_BRIGHTNESS_2 "\xee\x8e\xa7"
#define ICON_MD_BRIGHTNESS_3 "\xee\x8e\xa8"
#define ICON_MD_BRIGHTNESS_4 "\xee\x8e\xa9"
#define ICON_MD_BRIGHTNESS_5 "\xee\x8e\xaa"
#define ICON_MD_BRIGHTNESS_6 "\xee\x8e\xab"
#define ICON_MD_BRIGHTNESS_7 "\xee\x8e\xac"
#define ICON_MD_BRIGHTNESS_AUTO "\xee\x86\xab"
#define ICON_MD_BRIGHTNESS_HIGH "\xee\x86\xac"
#define ICON_MD_BRIGHTNESS_LOW "\xee\x86\xad"
#define ICON_MD_BRIGHTNESS_MEDIUM "\xee\x86\xae"
#define ICON_MD_BROKEN_IMAGE "\xee\x8e\xad"
#define ICON_MD_BRUSH "\xee\x8e\xae"
#define ICON_MD_BUBBLE_CHART "\xee\x9b\x9d"
#define ICON_MD_BUG_REPORT "\xee\xa1\xa8"
#define ICON_MD_BUILD "\xee\xa1\xa9"
#define ICON_MD_BURST_MODE "\xee\x90\xbc"
#define ICON_MD_BUSINESS "\xee\x82\xaf"
#define ICON_MD_BUSINESS_CENTER "\xee\xac\xbf"
#define ICON_MD_CACHED "\xee\xa1\xaa"
#define ICON_MD_CAKE "\xee\x9f\xa9"
#define ICON_MD_CALL "\xee\x82\xb0"
#define ICON_MD_CALL_END "\xee\x82\xb1"
#define ICON_MD_CALL_MADE "\xee\x82\xb2"
#define ICON_MD_CALL_MERGE "\xee\x82\xb3"
#define ICON_MD_CALL_MISSED "\xee\x82\xb4"
#define ICON_MD_CALL_MISSED_OUTGOING "\xee\x83\xa4"
#define ICON_MD_CALL_RECEIVED "\xee\x82\xb5"
#define ICON_MD_CALL_SPLIT "\xee\x82\xb6"
#define ICON_MD_CALL_TO_ACTION "\xee\x81\xac"
#define ICON_MD_CAMERA "\xee\x8e\xaf"
#define ICON_MD_CAMERA_ALT "\xee\x8e\xb0"
#define ICON_MD_CAMERA_ENHANCE "\xee\xa3\xbc"
#define ICON_MD_CAMERA_FRONT "\xee\x8e\xb1"
#define ICON_MD_CAMERA_REAR "\xee\x8e\xb2"
#define ICON_MD_CAMERA_ROLL "\xee\x8e\xb3"
#define ICON_MD_CANCEL "\xee\x97\x89"
#define ICON_MD_CARD_GIFTCARD "\xee\xa3\xb6"
#define ICON_MD_CARD_MEMBERSHIP "\xee\xa3\xb7"
#define ICON_MD_CARD_TRAVEL "\xee\xa3\xb8"
#define ICON_MD_CASINO "\xee\xad\x80"
#define ICON_MD_CAST "\xee\x8c\x87"
#define ICON_MD_CAST_CONNECTED "\xee\x8c\x88"
#define ICON_MD_CENTER_FOCUS_STRONG "\xee\x8e\xb4"
#define ICON_MD_CENTER_FOCUS_WEAK "\xee\x8e\xb5"
#define ICON_MD_CHANGE_HISTORY "\xee\xa1\xab"
#define ICON_MD_CHAT "\xee\x82\xb7"
#define ICON_MD_CHAT_BUBBLE "\xee\x83\x8a"
#define ICON_MD_CHAT_BUBBLE_OUTLINE "\xee\x83\x8b"
#define ICON_MD_CHECK "\xee\x97\x8a"
#define ICON_MD_CHECK_BOX "\xee\xa0\xb4"
#define ICON_MD_CHECK_BOX_OUTLINE_BLANK "\xee\xa0\xb5"
#define ICON_MD_CHECK_CIRCLE "\xee\xa1\xac"
#define ICON_MD_CHEVRON_LEFT "\xee\x97\x8b"
#define ICON_MD_CHEVRON_RIGHT "\xee\x97\x8c"
#define ICON_MD_CHILD_CARE "\xee\xad\x81"
#define ICON_MD_CHILD_FRIENDLY "\xee\xad\x82"
#define ICON_MD_CHROME_READER_MODE "\xee\xa1\xad"
#define ICON_MD_CLASS "\xee\xa1\xae"
#define ICON_MD_CLEAR "\xee\x85\x8c"
#define ICON_MD_CLEAR_ALL "\xee\x82\xb8"
#define ICON_MD_CLOSE "\xee\x97\x8d"
#define ICON_MD_CLOSED_CAPTION "\xee\x80\x9c"
#define ICON_MD_CLOUD "\xee\x8a\xbd"
#define ICON_MD_CLOUD_CIRCLE "\xee\x8a\xbe"
#define ICON_MD_CLOUD_DONE "\xee\x8a\xbf"
#define ICON_MD_CLOUD_DOWNLOAD "\xee\x8b\x80"
#define ICON_MD_CLOUD_OFF "\xee\x8b\x81"
#define ICON_MD_CLOUD_QUEUE "\xee\x8b\x82"
#define ICON_MD_CLOUD_UPLOAD "\xee\x8b\x83"
#define ICON_MD_CODE "\xee\xa1\xaf"
#define ICON_MD_COLLECTIONS "\xee\x8e\xb6"
#define ICON_MD_COLLECTIONS_BOOKMARK "\xee\x90\xb1"
#define ICON_MD_COLOR_LENS "\xee\x8e\xb7"
#define ICON_MD_COLORIZE "\xee\x8e\xb8"
#define ICON_MD_COMMENT "\xee\x82\xb9"
#define ICON_MD_COMPARE "\xee\x8e\xb9"
#define ICON_MD_COMPARE_ARROWS "\xee\xa4\x95"
#define ICON_MD_COMPUTER "\xee\x8c\x8a"
#define ICON_MD_CONFIRMATION_NUMBER "\xee\x98\xb8"
#define ICON_MD_CONTACT_MAIL "\xee\x83\x90"
#define ICON_MD_CONTACT_PHONE "\xee\x83\x8f"
#define ICON_MD_CONTACTS "\xee\x82\xba"
#define ICON_MD_CONTENT_COPY "\xee\x85\x8d"
#define ICON_MD_CONTENT_CUT "\xee\x85\x8e"
#define ICON_MD_CONTENT_PASTE "\xee\x85\x8f"
#define ICON_MD_CONTROL_POINT "\xee\x8e\xba"
#define ICON_MD_CONTROL_POINT_DUPLICATE "\xee\x8e\xbb"
#define ICON_MD_COPYRIGHT "\xee\xa4\x8c"
#define ICON_MD_CREATE "\xee\x85\x90"
#define ICON_MD_CREATE_NEW_FOLDER "\xee\x8b\x8c"
#define ICON_MD_CREDIT_CARD "\xee\xa1\xb0"
#define ICON_MD_CROP "\xee\x8e\xbe"
#define ICON_MD_CROP_16_9 "\xee\x8e\xbc"
#define ICON_MD_CROP_3_2 "\xee\x8e\xbd"
#define ICON_MD_CROP_5_4 "\xee\x8e\xbf"
#define ICON_MD_CROP_7_5 "\xee\x8f\x80"
#define ICON_MD_CROP_DIN "\xee\x8f\x81"
#define ICON_MD_CROP_FREE "\xee\x8f\x82"
#define ICON_MD_CROP_LANDSCAPE "\xee\x8f\x83"
#define ICON_MD_CROP_ORIGINAL "\xee\x8f\x84"
#define ICON_MD_CROP_PORTRAIT "\xee\x8f\x85"
#define ICON_MD_CROP_ROTATE "\xee\x90\xb7"
#define ICON_MD_CROP_SQUARE "\xee\x8f\x86"
#define ICON_MD_DASHBOARD "\xee\xa1\xb1"
#define ICON_MD_DATA_USAGE "\xee\x86\xaf"
#define ICON_MD_DATE_RANGE "\xee\xa4\x96"
#define ICON_MD_DEHAZE "\xee\x8f\x87"
#define ICON_MD_DELETE "\xee\xa1\xb2"
#define ICON_MD_DELETE_FOREVER "\xee\xa4\xab"
#define ICON_MD_DELETE_SWEEP "\xee\x85\xac"
#define ICON_MD_DESCRIPTION "\xee\xa1\xb3"
#define ICON_MD_DESKTOP_MAC "\xee\x8c\x8b"
#define ICON_MD_DESKTOP_WINDOWS "\xee\x8c\x8c"
#define ICON_MD_DETAILS "\xee\x8f\x88"
#define ICON_MD_DEVELOPER_BOARD "\xee\x8c\x8d"
#define ICON_MD_DEVELOPER_MODE "\xee\x86\xb0"
#define ICON_MD_DEVICE_HUB "\xee\x8c\xb5"
#define ICON_MD_DEVICES "\xee\x86\xb1"
#define ICON_MD_DEVICES_OTHER "\xee\x8c\xb7"
#define ICON_MD_DIALER_SIP "\xee\x82\xbb"
#define ICON_MD_DIALPAD "\xee\x82\xbc"
#define ICON_MD_DIRECTIONS "\xee\x94\xae"
#define ICON_MD_DIRECTIONS_BIKE "\xee\x94\xaf"
#define ICON_MD_DIRECTIONS_BOAT "\xee\x94\xb2"
#define ICON_MD_DIRECTIONS_BUS "\xee\x94\xb0"
#define ICON_MD_DIRECTIONS_CAR "\xee\x94\xb1"
#define ICON_MD_DIRECTIONS_RAILWAY "\xee\x94\xb4"
#define ICON_MD_DIRECTIONS_RUN "\xee\x95\xa6"
#define ICON_MD_DIRECTIONS_SUBWAY "\xee\x94\xb3"
#define ICON_MD_DIRECTIONS_TRANSIT "\xee\x94\xb5"
#define ICON_MD_DIRECTIONS_WALK "\xee\x94\xb6"
#define ICON_MD_DISC_FULL "\xee\x98\x90"
#define ICON_MD_DNS "\xee\xa1\xb5"
#define ICON_MD_DO_NOT_DISTURB "\xee\x98\x92"
#define ICON_MD_DO_NOT_DISTURB_ALT "\xee\x98\x91"
#define ICON_MD_DO_NOT_DISTURB_OFF "\xee\x99\x83"
#define ICON_MD_DO_NOT_DISTURB_ON "\xee\x99\x84"
#define ICON_MD_DOCK "\xee\x8c\x8e"
#define ICON_MD_DOMAIN "\xee\x9f\xae"
#define ICON_MD_DONE "\xee\xa1\xb6"
#define ICON_MD_DONE_ALL "\xee\xa1\xb7"
#define ICON_MD_DONUT_LARGE "\xee\xa4\x97"
#define ICON_MD_DONUT_SMALL "\xee\xa4\x98"
#define ICON_MD_DRAFTS "\xee\x85\x91"
#define ICON_MD_DRAG_HANDLE "\xee\x89\x9d"
#define ICON_MD_DRIVE_ETA "\xee\x98\x93"
#define ICON_MD_DVR "\xee\x86\xb2"
#define ICON_MD_EDIT "\xee\x8f\x89"
#define ICON_MD_EDIT_LOCATION "\xee\x95\xa8"
#define ICON_MD_EJECT "\xee\xa3\xbb"
#define ICON_MD_EMAIL "\xee\x82\xbe"
#define ICON_MD_ENHANCED_ENCRYPTION "\xee\x98\xbf"
#define ICON_MD_EQUALIZER "\xee\x80\x9d"
#define ICON_MD_ERROR "\xee\x80\x80"
#define ICON_MD_ERROR_OUTLINE "\xee\x80\x81"
#define ICON_MD_EURO_SYMBOL "\xee\xa4\xa6"
#define ICON_MD_EV_STATION "\xee\x95\xad"
#define ICON_MD_EVENT "\xee\xa1\xb8"
#define ICON_MD_EVENT_AVAILABLE "\xee\x98\x94"
#define ICON_MD_EVENT_BUSY "\xee\x98\x95"
#define ICON_MD_EVENT_NOTE "\xee\x98\x96"
#define ICON_MD_EVENT_SEAT "\xee\xa4\x83"
#define ICON_MD_EXIT_TO_APP "\xee\xa1\xb9"
#define ICON_MD_EXPAND_LESS "\xee\x97\x8e"
#define ICON_MD_EXPAND_MORE "\xee\x97\x8f"
#define ICON_MD_EXPLICIT "\xee\x80\x9e"
#define ICON_MD_EXPLORE "\xee\xa1\xba"
#define ICON_MD_EXPOSURE "\xee\x8f\x8a"
#define ICON_MD_EXPOSURE_NEG_1 "\xee\x8f\x8b"
#define ICON_MD_EXPOSURE_NEG_2 "\xee\x8f\x8c"
#define ICON_MD_EXPOSURE_PLUS_1 "\xee\x8f\x8d"
#define ICON_MD_EXPOSURE_PLUS_2 "\xee\x8f\x8e"
#define ICON_MD_EXPOSURE_ZERO "\xee\x8f\x8f"
#define ICON_MD_EXTENSION "\xee\xa1\xbb"
#define ICON_MD_FACE "\xee\xa1\xbc"
#define ICON_MD_FAST_FORWARD "\xee\x80\x9f"
#define ICON_MD_FAST_REWIND "\xee\x80\xa0"
#define ICON_MD_FAVORITE "\xee\xa1\xbd"
#define ICON_MD_FAVORITE_BORDER "\xee\xa1\xbe"
#define ICON_MD_FEATURED_PLAY_LIST "\xee\x81\xad"
#define ICON_MD_FEATURED_VIDEO "\xee\x81\xae"
#define ICON_MD_FEEDBACK "\xee\xa1\xbf"
#define ICON_MD_FIBER_DVR "\xee\x81\x9d"
#define ICON_MD_FIBER_MANUAL_RECORD "\xee\x81\xa1"
#define ICON_MD_FIBER_NEW "\xee\x81\x9e"
#define ICON_MD_FIBER_PIN "\xee\x81\xaa"
#define ICON_MD_FIBER_SMART_RECORD "\xee\x81\xa2"
#define ICON_MD_FILE_DOWNLOAD "\xee\x8b\x84"
#define ICON_MD_FILE_UPLOAD "\xee\x8b\x86"
#define ICON_MD_FILTER "\xee\x8f\x93"
#define ICON_MD_FILTER_1 "\xee\x8f\x90"
#define ICON_MD_FILTER_2 "\xee\x8f\x91"
#define ICON_MD_FILTER_3 "\xee\x8f\x92"
#define ICON_MD_FILTER_4 "\xee\x8f\x94"
#define ICON_MD_FILTER_5 "\xee\x8f\x95"
#define ICON_MD_FILTER_6 "\xee\x8f\x96"
#define ICON_MD_FILTER_7 "\xee\x8f\x97"
#define ICON_MD_FILTER_8 "\xee\x8f\x98"
#define ICON_MD_FILTER_9 "\xee\x8f\x99"
#define ICON_MD_FILTER_9_PLUS "\xee\x8f\x9a"
#define ICON_MD_FILTER_B_AND_W "\xee\x8f\x9b"
#define ICON_MD_FILTER_CENTER_FOCUS "\xee\x8f\x9c"
#define ICON_MD_FILTER_DRAMA "\xee\x8f\x9d"
#define ICON_MD_FILTER_FRAMES "\xee\x8f\x9e"
#define ICON_MD_FILTER_HDR "\xee\x8f\x9f"
#define ICON_MD_FILTER_LIST "\xee\x85\x92"
#define ICON_MD_FILTER_NONE "\xee\x8f\xa0"
#define ICON_MD_FILTER_TILT_SHIFT "\xee\x8f\xa2"
#define ICON_MD_FILTER_VINTAGE "\xee\x8f\xa3"
#define ICON_MD_FIND_IN_PAGE "\xee\xa2\x80"
#define ICON_MD_FIND_REPLACE "\xee\xa2\x81"
#define ICON_MD_FINGERPRINT "\xee\xa4\x8d"
#define ICON_MD_FIRST_PAGE "\xee\x97\x9c"
#define ICON_MD_FITNESS_CENTER "\xee\xad\x83"
#define ICON_MD_FLAG "\xee\x85\x93"
#define ICON_MD_FLARE "\xee\x8f\xa4"
#define ICON_MD_FLASH_AUTO "\xee\x8f\xa5"
#define ICON_MD_FLASH_OFF "\xee\x8f\xa6"
#define ICON_MD_FLASH_ON "\xee\x8f\xa7"
#define ICON_MD_FLIGHT "\xee\x94\xb9"
#define ICON_MD_FLIGHT_LAND "\xee\xa4\x84"
#define ICON_MD_FLIGHT_TAKEOFF "\xee\xa4\x85"
#define ICON_MD_FLIP "\xee\x8f\xa8"
#define ICON_MD_FLIP_TO_BACK "\xee\xa2\x82"
#define ICON_MD_FLIP_TO_FRONT "\xee\xa2\x83"
#define ICON_MD_FOLDER "\xee\x8b\x87"
#define ICON_MD_FOLDER_OPEN "\xee\x8b\x88"
#define ICON_MD_FOLDER_SHARED "\xee\x8b\x89"
#define ICON_MD_FOLDER_SPECIAL "\xee\x98\x97"
#define ICON_MD_FONT_DOWNLOAD "\xee\x85\xa7"
#define ICON_MD_FORMAT_ALIGN_CENTER "\xee\x88\xb4"
#define ICON_MD_FORMAT_ALIGN_JUSTIFY "\xee\x88\xb5"
#define ICON_MD_FORMAT_ALIGN_LEFT "\xee\x88\xb6"
#define ICON_MD_FORMAT_ALIGN_RIGHT "\xee\x88\xb7"
#define ICON_MD_FORMAT_BOLD "\xee\x88\xb8"
#define ICON_MD_FORMAT_CLEAR "\xee\x88\xb9"
#define ICON_MD_FORMAT_COLOR_FILL "\xee\x88\xba"
#define ICON_MD_FORMAT_COLOR_RESET "\xee\x88\xbb"
#define ICON_MD_FORMAT_COLOR_TEXT "\xee\x88\xbc"
#define ICON_MD_FORMAT_INDENT_DECREASE "\xee\x88\xbd"
#define ICON_MD_FORMAT_INDENT_INCREASE "\xee\x88\xbe"
#define ICON_MD_FORMAT_ITALIC "\xee\x88\xbf"
#define ICON_MD_FORMAT_LINE_SPACING "\xee\x89\x80"
#define ICON_MD_FORMAT_LIST_BULLETED "\xee\x89\x81"
#define ICON_MD_FORMAT_LIST_NUMBERED "\xee\x89\x82"
#define ICON_MD_FORMAT_PAINT "\xee\x89\x83"
#define ICON_MD_FORMAT_QUOTE "\xee\x89\x84"
#define ICON_MD_FORMAT_SHAPES "\xee\x89\x9e"
#define ICON_MD_FORMAT_SIZE "\xee\x89\x85"
#define ICON_MD_FORMAT_STRIKETHROUGH "\xee\x89\x86"
#define ICON_MD_FORMAT_TEXTDIRECTION_L_TO_R "\xee\x89\x87"
#define ICON_MD_FORMAT_TEXTDIRECTION_R_TO_L "\xee\x89\x88"
#define ICON_MD_FORMAT_UNDERLINED "\xee\x89\x89"
#define ICON_MD_FORUM "\xee\x82\xbf"
#define ICON_MD_FORWARD "\xee\x85\x94"
#define ICON_MD_FORWARD_10 "\xee\x81\x96"
#define ICON_MD_FORWARD_30 "\xee\x81\x97"
#define ICON_MD_FORWARD_5 "\xee\x81\x98"
#define ICON_MD_FREE_BREAKFAST "\xee\xad\x84"
#define ICON_MD_FULLSCREEN "\xee\x97\x90"
#define ICON_MD_FULLSCREEN_EXIT "\xee\x97\x91"
#define ICON_MD_FUNCTIONS "\xee\x89\x8a"
#define ICON_MD_G_TRANSLATE "\xee\xa4\xa7"
#define ICON_MD_GAMEPAD "\xee\x8c\x8f"
#define ICON_MD_GAMES "\xee\x80\xa1"
#define ICON_MD_GAVEL "\xee\xa4\x8e"
#define ICON_MD_GESTURE "\xee\x85\x95"
#define ICON_MD_GET_APP "\xee\xa2\x84"
#define ICON_MD_GIF "\xee\xa4\x88"
#define ICON_MD_GOLF_COURSE "\xee\xad\x85"
#define ICON_MD_GPS_FIXED "\xee\x86\xb3"
#define ICON_MD_GPS_NOT_FIXED "\xee\x86\xb4"
#define ICON_MD_GPS_OFF "\xee\x86\xb5"
#define ICON_MD_GRADE "\xee\xa2\x85"
#define ICON_MD_GRADIENT "\xee\x8f\xa9"
#define ICON_MD_GRAIN "\xee\x8f\xaa"
#define ICON_MD_GRAPHIC_EQ "\xee\x86\xb8"
#define ICON_MD_GRID_OFF "\xee\x8f\xab"
#define ICON_MD_GRID_ON "\xee\x8f\xac"
#define ICON_MD_GROUP "\xee\x9f\xaf"
#define ICON_MD_GROUP_ADD "\xee\x9f\xb0"
#define ICON_MD_GROUP_WORK "\xee\xa2\x86"
#define ICON_MD_HD "\xee\x81\x92"
#define ICON_MD_HDR_OFF "\xee\x8f\xad"
#define ICON_MD_HDR_ON "\xee\x8f\xae"
#define ICON_MD_HDR_STRONG "\xee\x8f\xb1"
#define ICON_MD_HDR_WEAK "\xee\x8f\xb2"
#define ICON_MD_HEADSET "\xee\x8c\x90"
#define ICON_MD_HEADSET_MIC "\xee\x8c\x91"
#define ICON_MD_HEALING "\xee\x8f\xb3"
#define ICON_MD_HEARING "\xee\x80\xa3"
#define ICON_MD_HELP "\xee\xa2\x87"
#define ICON_MD_HELP_OUTLINE "\xee\xa3\xbd"
#define ICON_MD_HIGH_QUALITY "\xee\x80\xa4"
#define ICON_MD_HIGHLIGHT "\xee\x89\x9f"
#define ICON_MD_HIGHLIGHT_OFF "\xee\xa2\x88"
#define ICON_MD_HISTORY "\xee\xa2\x89"
#define ICON_MD_HOME "\xee\xa2\x8a"
#define ICON_MD_HOT_TUB "\xee\xad\x86"
#define ICON_MD_HOTEL "\xee\x94\xba"
#define ICON_MD_HOURGLASS_EMPTY "\xee\xa2\x8b"
#define ICON_MD_HOURGLASS_FULL "\xee\xa2\x8c"
#define ICON_MD_HTTP "\xee\xa4\x82"
#define ICON_MD_HTTPS "\xee\xa2\x8d"
#define ICON_MD_IMAGE "\xee\x8f\xb4"
#define ICON_MD_IMAGE_ASPECT_RATIO "\xee\x8f\xb5"
#define ICON_MD_IMPORT_CONTACTS "\xee\x83\xa0"
#define ICON_MD_IMPORT_EXPORT "\xee\x83\x83"
#define ICON_MD_IMPORTANT_DEVICES "\xee\xa4\x92"
#define ICON_MD_INBOX "\xee\x85\x96"
#define ICON_MD_INDETERMINATE_CHECK_BOX "\xee\xa4\x89"
#define ICON_MD_INFO "\xee\xa2\x8e"
#define ICON_MD_INFO_OUTLINE "\xee\xa2\x8f"
#define ICON_MD_INPUT "\xee\xa2\x90"
#define ICON_MD_INSERT_CHART "\xee\x89\x8b"
#define ICON_MD_INSERT_COMMENT "\xee\x89\x8c"
#define ICON_MD_INSERT_DRIVE_FILE "\xee\x89\x8d"
#define ICON_MD_INSERT_EMOTICON "\xee\x89\x8e"
#define ICON_MD_INSERT_INVITATION "\xee\x89\x8f"
#define ICON_MD_INSERT_LINK "\xee\x89\x90"
#define ICON_MD_INSERT_PHOTO "\xee\x89\x91"
#define ICON_MD_INVERT_COLORS "\xee\xa2\x91"
#define ICON_MD_INVERT_COLORS_OFF "\xee\x83\x84"
#define ICON_MD_ISO "\xee\x8f\xb6"
#define ICON_MD_KEYBOARD "\xee\x8c\x92"
#define ICON_MD_KEYBOARD_ARROW_DOWN "\xee\x8c\x93"
#define ICON_MD_KEYBOARD_ARROW_LEFT "\xee\x8c\x94"
#define ICON_MD_KEYBOARD_ARROW_RIGHT "\xee\x8c\x95"
#define ICON_MD_KEYBOARD_ARROW_UP "\xee\x8c\x96"
#define ICON_MD_KEYBOARD_BACKSPACE "\xee\x8c\x97"
#define ICON_MD_KEYBOARD_CAPSLOCK "\xee\x8c\x98"
#define ICON_MD_KEYBOARD_HIDE "\xee\x8c\x9a"
#define ICON_MD_KEYBOARD_RETURN "\xee\x8c\x9b"
#define ICON_MD_KEYBOARD_TAB "\xee\x8c\x9c"
#define ICON_MD_KEYBOARD_VOICE "\xee\x8c\x9d"
#define ICON_MD_KITCHEN "\xee\xad\x87"
#define ICON_MD_LABEL "\xee\xa2\x92"
#define ICON_MD_LABEL_OUTLINE "\xee\xa2\x93"
#define ICON_MD_LANDSCAPE "\xee\x8f\xb7"
#define ICON_MD_LANGUAGE "\xee\xa2\x94"
#define ICON_MD_LAPTOP "\xee\x8c\x9e"
#define ICON_MD_LAPTOP_CHROMEBOOK "\xee\x8c\x9f"
#define ICON_MD_LAPTOP_MAC "\xee\x8c\xa0"
#define ICON_MD_LAPTOP_WINDOWS "\xee\x8c\xa1"
#define ICON_MD_LAST_PAGE "\xee\x97\x9d"
#define ICON_MD_LAUNCH "\xee\xa2\x95"
#define ICON_MD_LAYERS "\xee\x94\xbb"
#define ICON_MD_LAYERS_CLEAR "\xee\x94\xbc"
#define ICON_MD_LEAK_ADD "\xee\x8f\xb8"
#define ICON_MD_LEAK_REMOVE "\xee\x8f\xb9"
#define ICON_MD_LENS "\xee\x8f\xba"
#define ICON_MD_LIBRARY_ADD "\xee\x80\xae"
#define ICON_MD_LIBRARY_BOOKS "\xee\x80\xaf"
#define ICON_MD_LIBRARY_MUSIC "\xee\x80\xb0"
#define ICON_MD_LIGHTBULB_OUTLINE "\xee\xa4\x8f"
#define ICON_MD_LINE_STYLE "\xee\xa4\x99"
#define ICON_MD_LINE_WEIGHT "\xee\xa4\x9a"
#define ICON_MD_LINEAR_SCALE "\xee\x89\xa0"
#define ICON_MD_LINK "\xee\x85\x97"
#define ICON_MD_LINKED_CAMERA "\xee\x90\xb8"
#define ICON_MD_LIST "\xee\xa2\x96"
#define ICON_MD_LIVE_HELP "\xee\x83\x86"
#define ICON_MD_LIVE_TV "\xee\x98\xb9"
#define ICON_MD_LOCAL_ACTIVITY "\xee\x94\xbf"
#define ICON_MD_LOCAL_AIRPORT "\xee\x94\xbd"
#define ICON_MD_LOCAL_ATM "\xee\x94\xbe"
#define ICON_MD_LOCAL_BAR "\xee\x95\x80"
#define ICON_MD_LOCAL_CAFE "\xee\x95\x81"
#define ICON_MD_LOCAL_CAR_WASH "\xee\x95\x82"
#define ICON_MD_LOCAL_CONVENIENCE_STORE "\xee\x95\x83"
#define ICON_MD_LOCAL_DINING "\xee\x95\x96"
#define ICON_MD_LOCAL_DRINK "\xee\x95\x84"
#define ICON_MD_LOCAL_FLORIST "\xee\x95\x85"
#define ICON_MD_LOCAL_GAS_STATION "\xee\x95\x86"
#define ICON_MD_LOCAL_GROCERY_STORE "\xee\x95\x87"
#define ICON_MD_LOCAL_HOSPITAL "\xee\x95\x88"
#define ICON_MD_LOCAL_HOTEL "\xee\x95\x89"
#define ICON_MD_LOCAL_LAUNDRY_SERVICE "\xee\x95\x8a"
#define ICON_MD_LOCAL_LIBRARY "\xee\x95\x8b"
#define ICON_MD_LOCAL_MALL "\xee\x95\x8c"
#define ICON_MD_LOCAL_MOVIES "\xee\x95\x8d"
#define ICON_MD_LOCAL_OFFER "\xee\x95\x8e"
#define ICON_MD_LOCAL_PARKING "\xee\x95\x8f"
#define ICON_MD_LOCAL_PHARMACY "\xee\x95\x90"
#define ICON_MD_LOCAL_PHONE "\xee\x95\x91"
#define ICON_MD_LOCAL_PIZZA "\xee\x95\x92"
#define ICON_MD_LOCAL_PLAY "\xee\x95\x93"
#define ICON_MD_LOCAL_POST_OFFICE "\xee\x95\x94"
#define ICON_MD_LOCAL_PRINTSHOP "\xee\x95\x95"
#define ICON_MD_LOCAL_SEE "\xee\x95\x97"
#define ICON_MD_LOCAL_SHIPPING "\xee\x95\x98"
#define ICON_MD_LOCAL_TAXI "\xee\x95\x99"
#define ICON_MD_LOCATION_CITY "\xee\x9f\xb1"
#define ICON_MD_LOCATION_DISABLED "\xee\x86\xb6"
#define ICON_MD_LOCATION_OFF "\xee\x83\x87"
#define ICON_MD_LOCATION_ON "\xee\x83\x88"
#define ICON_MD_LOCATION_SEARCHING "\xee\x86\xb7"
#define ICON_MD_LOCK "\xee\xa2\x97"
#define ICON_MD_LOCK_OPEN "\xee\xa2\x98"
#define ICON_MD_LOCK_OUTLINE "\xee\xa2\x99"
#define ICON_MD_LOOKS "\xee\x8f\xbc"
#define ICON_MD_LOOKS_3 "\xee\x8f\xbb"
#define ICON_MD_LOOKS_4 "\xee\x8f\xbd"
#define ICON_MD_LOOKS_5 "\xee\x8f\xbe"
#define ICON_MD_LOOKS_6 "\xee\x8f\xbf"
#define ICON_MD_LOOKS_ONE "\xee\x90\x80"
#define ICON_MD_LOOKS_TWO "\xee\x90\x81"
#define ICON_MD_LOOP "\xee\x80\xa8"
#define ICON_MD_LOUPE "\xee\x90\x82"
#define ICON_MD_LOW_PRIORITY "\xee\x85\xad"
#define ICON_MD_LOYALTY "\xee\xa2\x9a"
#define ICON_MD_MAIL "\xee\x85\x98"
#define ICON_MD_MAIL_OUTLINE "\xee\x83\xa1"
#define ICON_MD_MAP "\xee\x95\x9b"
#define ICON_MD_MARKUNREAD "\xee\x85\x99"
#define ICON_MD_MARKUNREAD_MAILBOX "\xee\xa2\x9b"
#define ICON_MD_MEMORY "\xee\x8c\xa2"
#define ICON_MD_MENU "\xee\x97\x92"
#define ICON_MD_MERGE_TYPE "\xee\x89\x92"
#define ICON_MD_MESSAGE "\xee\x83\x89"
#define ICON_MD_MIC "\xee\x80\xa9"
#define ICON_MD_MIC_NONE "\xee\x80\xaa"
#define ICON_MD_MIC_OFF "\xee\x80\xab"
#define ICON_MD_MMS "\xee\x98\x98"
#define ICON_MD_MODE_COMMENT "\xee\x89\x93"
#define ICON_MD_MODE_EDIT "\xee\x89\x94"
#define ICON_MD_MONETIZATION_ON "\xee\x89\xa3"
#define ICON_MD_MONEY_OFF "\xee\x89\x9c"
#define ICON_MD_MONOCHROME_PHOTOS "\xee\x90\x83"
#define ICON_MD_MOOD "\xee\x9f\xb2"
#define ICON_MD_MOOD_BAD "\xee\x9f\xb3"
#define ICON_MD_MORE "\xee\x98\x99"
#define ICON_MD_MORE_HORIZ "\xee\x97\x93"
#define ICON_MD_MORE_VERT "\xee\x97\x94"
#define ICON_MD_MOTORCYCLE "\xee\xa4\x9b"
#define ICON_MD_MOUSE "\xee\x8c\xa3"
#define ICON_MD_MOVE_TO_INBOX "\xee\x85\xa8"
#define ICON_MD_MOVIE "\xee\x80\xac"
#define ICON_MD_MOVIE_CREATION "\xee\x90\x84"
#define ICON_MD_MOVIE_FILTER "\xee\x90\xba"
#define ICON_MD_MULTILINE_CHART "\xee\x9b\x9f"
#define ICON_MD_MUSIC_NOTE "\xee\x90\x85"
#define ICON_MD_MUSIC_VIDEO "\xee\x81\xa3"
#define ICON_MD_MY_LOCATION "\xee\x95\x9c"
#define ICON_MD_NATURE "\xee\x90\x86"
#define ICON_MD_NATURE_PEOPLE "\xee\x90\x87"
#define ICON_MD_NAVIGATE_BEFORE "\xee\x90\x88"
#define ICON_MD_NAVIGATE_NEXT "\xee\x90\x89"
#define ICON_MD_NAVIGATION "\xee\x95\x9d"
#define ICON_MD_NEAR_ME "\xee\x95\xa9"
#define ICON_MD_NETWORK_CELL "\xee\x86\xb9"
#define ICON_MD_NETWORK_CHECK "\xee\x99\x80"
#define ICON_MD_NETWORK_LOCKED "\xee\x98\x9a"
#define ICON_MD_NETWORK_WIFI "\xee\x86\xba"
#define ICON_MD_NEW_RELEASES "\xee\x80\xb1"
#define ICON_MD_NEXT_WEEK "\xee\x85\xaa"
#define ICON_MD_NFC "\xee\x86\xbb"
#define ICON_MD_NO_ENCRYPTION "\xee\x99\x81"
#define ICON_MD_NO_SIM "\xee\x83\x8c"
#define ICON_MD_NOT_INTERESTED "\xee\x80\xb3"
#define ICON_MD_NOTE "\xee\x81\xaf"
#define ICON_MD_NOTE_ADD "\xee\xa2\x9c"
#define ICON_MD_NOTIFICATIONS "\xee\x9f\xb4"
#define ICON_MD_NOTIFICATIONS_ACTIVE "\xee\x9f\xb7"
#define ICON_MD_NOTIFICATIONS_NONE "\xee\x9f\xb5"
#define ICON_MD_NOTIFICATIONS_OFF "\xee\x9f\xb6"
#define ICON_MD_NOTIFICATIONS_PAUSED "\xee\x9f\xb8"
#define ICON_MD_OFFLINE_PIN "\xee\xa4\x8a"
#define ICON_MD_ONDEMAND_VIDEO "\xee\x98\xba"
#define ICON_MD_OPACITY "\xee\xa4\x9c"
#define ICON_MD_OPEN_IN_BROWSER "\xee\xa2\x9d"
#define ICON_MD_OPEN_IN_NEW "\xee\xa2\x9e"
#define ICON_MD_OPEN_WITH "\xee\xa2\x9f"
#define ICON_MD_PAGES "\xee\x9f\xb9"
#define ICON_MD_PAGEVIEW "\xee\xa2\xa0"
#define ICON_MD_PALETTE "\xee\x90\x8a"
#define ICON_MD_PAN_TOOL "\xee\xa4\xa5"
#define ICON_MD_PANORAMA "\xee\x90\x8b"
#define ICON_MD_PANORAMA_FISH_EYE "\xee\x90\x8c"
#define ICON_MD_PANORAMA_HORIZONTAL "\xee\x90\x8d"
#define ICON_MD_PANORAMA_VERTICAL "\xee\x90\x8e"
#define ICON_MD_PANORAMA_WIDE_ANGLE "\xee\x90\x8f"
#define ICON_MD_PARTY_MODE "\xee\x9f\xba"
#define ICON_MD_PAUSE "\xee\x80\xb4"
#define ICON_MD_PAUSE_CIRCLE_FILLED "\xee\x80\xb5"
#define ICON_MD_PAUSE_CIRCLE_OUTLINE "\xee\x80\xb6"
#define ICON_MD_PAYMENT "\xee\xa2\xa1"
#define ICON_MD_PEOPLE "\xee\x9f\xbb"
#define ICON_MD_PEOPLE_OUTLINE "\xee\x9f\xbc"
#define ICON_MD_PERM_CAMERA_MIC "\xee\xa2\xa2"
#define ICON_MD_PERM_CONTACT_CALENDAR "\xee\xa2\xa3"
#define ICON_MD_PERM_DATA_SETTING "\xee\xa2\xa4"
#define ICON_MD_PERM_DEVICE_INFORMATION "\xee\xa2\xa5"
#define ICON_MD_PERM_IDENTITY "\xee\xa2\xa6"
#define ICON_MD_PERM_MEDIA "\xee\xa2\xa7"
#define ICON_MD_PERM_PHONE_MSG "\xee\xa2\xa8"
#define ICON_MD_PERM_SCAN_WIFI "\xee\xa2\xa9"
#define ICON_MD_PERSON "\xee\x9f\xbd"
#define ICON_MD_PERSON_ADD "\xee\x9f\xbe"
#define ICON_MD_PERSON_OUTLINE "\xee\x9f\xbf"
#define ICON_MD_PERSON_PIN "\xee\x95\x9a"
#define ICON_MD_PERSON_PIN_CIRCLE "\xee\x95\xaa"
#define ICON_MD_PERSONAL_VIDEO "\xee\x98\xbb"
#define ICON_MD_PETS "\xee\xa4\x9d"
#define ICON_MD_PHONE "\xee\x83\x8d"
#define ICON_MD_PHONE_ANDROID "\xee\x8c\xa4"
#define ICON_MD_PHONE_BLUETOOTH_SPEAKER "\xee\x98\x9b"
#define ICON_MD_PHONE_FORWARDED "\xee\x98\x9c"
#define ICON_MD_PHONE_IN_TALK "\xee\x98\x9d"
#define ICON_MD_PHONE_IPHONE "\xee\x8c\xa5"
#define ICON_MD_PHONE_LOCKED "\xee\x98\x9e"
#define ICON_MD_PHONE_MISSED "\xee\x98\x9f"
#define ICON_MD_PHONE_PAUSED "\xee\x98\xa0"
#define ICON_MD_PHONELINK "\xee\x8c\xa6"
#define ICON_MD_PHONELINK_ERASE "\xee\x83\x9b"
#define ICON_MD_PHONELINK_LOCK "\xee\x83\x9c"
#define ICON_MD_PHONELINK_OFF "\xee\x8c\xa7"
#define ICON_MD_PHONELINK_RING "\xee\x83\x9d"
#define ICON_MD_PHONELINK_SETUP "\xee\x83\x9e"
#define ICON_MD_PHOTO "\xee\x90\x90"
#define ICON_MD_PHOTO_ALBUM "\xee\x90\x91"
#define ICON_MD_PHOTO_CAMERA "\xee\x90\x92"
#define ICON_MD_PHOTO_FILTER "\xee\x90\xbb"
#define ICON_MD_PHOTO_LIBRARY "\xee\x90\x93"
#define ICON_MD_PHOTO_SIZE_SELECT_ACTUAL "\xee\x90\xb2"
#define ICON_MD_PHOTO_SIZE_SELECT_LARGE "\xee\x90\xb3"
#define ICON_MD_PHOTO_SIZE_SELECT_SMALL "\xee\x90\xb4"
#define ICON_MD_PICTURE_AS_PDF "\xee\x90\x95"
#define ICON_MD_PICTURE_IN_PICTURE "\xee\xa2\xaa"
#define ICON_MD_PICTURE_IN_PICTURE_ALT "\xee\xa4\x91"
#define ICON_MD_PIE_CHART "\xee\x9b\x84"
#define ICON_MD_PIE_CHART_OUTLINED "\xee\x9b\x85"
#define ICON_MD_PIN_DROP "\xee\x95\x9e"
#define ICON_MD_PLACE "\xee\x95\x9f"
#define ICON_MD_PLAY_ARROW "\xee\x80\xb7"
#define ICON_MD_PLAY_CIRCLE_FILLED "\xee\x80\xb8"
#define ICON_MD_PLAY_CIRCLE_OUTLINE "\xee\x80\xb9"
#define ICON_MD_PLAY_FOR_WORK "\xee\xa4\x86"
#define ICON_MD_PLAYLIST_ADD "\xee\x80\xbb"
#define ICON_MD_PLAYLIST_ADD_CHECK "\xee\x81\xa5"
#define ICON_MD_PLAYLIST_PLAY "\xee\x81\x9f"
#define ICON_MD_PLUS_ONE "\xee\xa0\x80"
#define ICON_MD_POLL "\xee\xa0\x81"
#define ICON_MD_POLYMER "\xee\xa2\xab"
#define ICON_MD_POOL "\xee\xad\x88"
#define ICON_MD_PORTABLE_WIFI_OFF "\xee\x83\x8e"
#define ICON_MD_PORTRAIT "\xee\x90\x96"
#define ICON_MD_POWER "\xee\x98\xbc"
#define ICON_MD_POWER_INPUT "\xee\x8c\xb6"
#define ICON_MD_POWER_SETTINGS_NEW "\xee\xa2\xac"
#define ICON_MD_PREGNANT_WOMAN "\xee\xa4\x9e"
#define ICON_MD_PRESENT_TO_ALL "\xee\x83\x9f"
#define ICON_MD_PRINT "\xee\xa2\xad"
#define ICON_MD_PRIORITY_HIGH "\xee\x99\x85"
#define ICON_MD_PUBLIC "\xee\xa0\x8b"
#define ICON_MD_PUBLISH "\xee\x89\x95"
#define ICON_MD_QUERY_BUILDER "\xee\xa2\xae"
#define ICON_MD_QUESTION_ANSWER "\xee\xa2\xaf"
#define ICON_MD_QUEUE "\xee\x80\xbc"
#define ICON_MD_QUEUE_MUSIC "\xee\x80\xbd"
#define ICON_MD_QUEUE_PLAY_NEXT "\xee\x81\xa6"
#define ICON_MD_RADIO "\xee\x80\xbe"
#define ICON_MD_RADIO_BUTTON_CHECKED "\xee\xa0\xb7"
#define ICON_MD_RADIO_BUTTON_UNCHECKED "\xee\xa0\xb6"
#define ICON_MD_RATE_REVIEW "\xee\x95\xa0"
#define ICON_MD_RECEIPT "\xee\xa2\xb0"
#define ICON_MD_RECENT_ACTORS "\xee\x80\xbf"
#define ICON_MD_RECORD_VOICE_OVER "\xee\xa4\x9f"
#define ICON_MD_REDEEM "\xee\xa2\xb1"
#define ICON_MD_REDO "\xee\x85\x9a"
#define ICON_MD_REFRESH "\xee\x97\x95"
#define ICON_MD_REMOVE "\xee\x85\x9b"
#define ICON_MD_REMOVE_CIRCLE "\xee\x85\x9c"
#define ICON_MD_REMOVE_CIRCLE_OUTLINE "\xee\x85\x9d"
#define ICON_MD_REMOVE_FROM_QUEUE "\xee\x81\xa7"
#define ICON_MD_REMOVE_RED_EYE "\xee\x90\x97"
#define ICON_MD_REMOVE_SHOPPING_CART "\xee\xa4\xa8"
#define ICON_MD_REORDER "\xee\xa3\xbe"
#define ICON_MD_REPEAT "\xee\x81\x80"
#define ICON_MD_REPEAT_ONE "\xee\x81\x81"
#define ICON_MD_REPLAY "\xee\x81\x82"
#define ICON_MD_REPLAY_10 "\xee\x81\x99"
#define ICON_MD_REPLAY_30 "\xee\x81\x9a"
#define ICON_MD_REPLAY_5 "\xee\x81\x9b"
#define ICON_MD_REPLY "\xee\x85\x9e"
#define ICON_MD_REPLY_ALL "\xee\x85\x9f"
#define ICON_MD_REPORT "\xee\x85\xa0"
#define ICON_MD_REPORT_PROBLEM "\xee\xa2\xb2"
#define ICON_MD_RESTAURANT "\xee\x95\xac"
#define ICON_MD_RESTAURANT_MENU "\xee\x95\xa1"
#define ICON_MD_RESTORE "\xee\xa2\xb3"
#define ICON_MD_RESTORE_PAGE "\xee\xa4\xa9"
#define ICON_MD_RING_VOLUME "\xee\x83\x91"
#define ICON_MD_ROOM "\xee\xa2\xb4"
#define ICON_MD_ROOM_SERVICE "\xee\xad\x89"
#define ICON_MD_ROTATE_90_DEGREES_CCW "\xee\x90\x98"
#define ICON_MD_ROTATE_LEFT "\xee\x90\x99"
#define ICON_MD_ROTATE_RIGHT "\xee\x90\x9a"
#define ICON_MD_ROUNDED_CORNER "\xee\xa4\xa0"
#define ICON_MD_ROUTER "\xee\x8c\xa8"
#define ICON_MD_ROWING "\xee\xa4\xa1"
#define ICON_MD_RSS_FEED "\xee\x83\xa5"
#define ICON_MD_RV_HOOKUP "\xee\x99\x82"
#define ICON_MD_SATELLITE "\xee\x95\xa2"
#define ICON_MD_SAVE "\xee\x85\xa1"
#define ICON_MD_SCANNER "\xee\x8c\xa9"
#define ICON_MD_SCHEDULE "\xee\xa2\xb5"
#define ICON_MD_SCHOOL "\xee\xa0\x8c"
#define ICON_MD_SCREEN_LOCK_LANDSCAPE "\xee\x86\xbe"
#define ICON_MD_SCREEN_LOCK_PORTRAIT "\xee\x86\xbf"
#define ICON_MD_SCREEN_LOCK_ROTATION "\xee\x87\x80"
#define ICON_MD_SCREEN_ROTATION "\xee\x87\x81"
#define ICON_MD_SCREEN_SHARE "\xee\x83\xa2"
#define ICON_MD_SD_CARD "\xee\x98\xa3"
#define ICON_MD_SD_STORAGE "\xee\x87\x82"
#define ICON_MD_SEARCH "\xee\xa2\xb6"
#define ICON_MD_SECURITY "\xee\x8c\xaa"
#define ICON_MD_SELECT_ALL "\xee\x85\xa2"
#define ICON_MD_SEND "\xee\x85\xa3"
#define ICON_MD_SENTIMENT_DISSATISFIED "\xee\xa0\x91"
#define ICON_MD_SENTIMENT_NEUTRAL "\xee\xa0\x92"
#define ICON_MD_SENTIMENT_SATISFIED "\xee\xa0\x93"
#define ICON_MD_SENTIMENT_VERY_DISSATISFIED "\xee\xa0\x94"
#define ICON_MD_SENTIMENT_VERY_SATISFIED "\xee\xa0\x95"
#define ICON_MD_SETTINGS "\xee\xa2\xb8"
#define ICON_MD_SETTINGS_APPLICATIONS "\xee\xa2\xb9"
#define ICON_MD_SETTINGS_BACKUP_RESTORE "\xee\xa2\xba"
#define ICON_MD_SETTINGS_BLUETOOTH "\xee\xa2\xbb"
#define ICON_MD_SETTINGS_BRIGHTNESS "\xee\xa2\xbd"
#define ICON_MD_SETTINGS_CELL "\xee\xa2\xbc"
#define ICON_MD_SETTINGS_ETHERNET "\xee\xa2\xbe"
#define ICON_MD_SETTINGS_INPUT_ANTENNA "\xee\xa2\xbf"
#define ICON_MD_SETTINGS_INPUT_COMPONENT "\xee\xa3\x80"
#define ICON_MD_SETTINGS_INPUT_COMPOSITE "\xee\xa3\x81"
#define ICON_MD_SETTINGS_INPUT_HDMI "\xee\xa3\x82"
#define ICON_MD_SETTINGS_INPUT_SVIDEO "\xee\xa3\x83"
#define ICON_MD_SETTINGS_OVERSCAN "\xee\xa3\x84"
#define ICON_MD_SETTINGS_PHONE "\xee\xa3\x85"
#define ICON_MD_SETTINGS_POWER "\xee\xa3\x86"
#define ICON_MD_SETTINGS_REMOTE "\xee\xa3\x87"
#define ICON_MD_SETTINGS_SYSTEM_DAYDREAM "\xee\x87\x83"
#define ICON_MD_SETTINGS_VOICE "\xee\xa3\x88"
#define ICON_MD_SHARE "\xee\xa0\x8d"
#define ICON_MD_SHOP "\xee\xa3\x89"
#define ICON_MD_SHOP_TWO "\xee\xa3\x8a"
#define ICON_MD_SHOPPING_BASKET "\xee\xa3\x8b"
#define ICON_MD_SHOPPING_CART "\xee\xa3\x8c"
#define ICON_MD_SHORT_TEXT "\xee\x89\xa1"
#define ICON_MD_SHOW_CHART "\xee\x9b\xa1"
#define ICON_MD_SHUFFLE "\xee\x81\x83"
#define ICON_MD_SIGNAL_CELLULAR_4_BAR "\xee\x87\x88"
#define ICON_MD_SIGNAL_CELLULAR_CONNECTED_NO_INTERNET_4_BAR "\xee\x87\x8d"
#define ICON_MD_SIGNAL_CELLULAR_NO_SIM "\xee\x87\x8e"
#define ICON_MD_SIGNAL_CELLULAR_NULL "\xee\x87\x8f"
#define ICON_MD_SIGNAL_CELLULAR_OFF "\xee\x87\x90"
#define ICON_MD_SIGNAL_WIFI_4_BAR "\xee\x87\x98"
#define ICON_MD_SIGNAL_WIFI_4_BAR_LOCK "\xee\x87\x99"
#define ICON_MD_SIGNAL_WIFI_OFF "\xee\x87\x9a"
#define ICON_MD_SIM_CARD "\xee\x8c\xab"
#define ICON_MD_SIM_CARD_ALERT "\xee\x98\xa4"
#define ICON_MD_SKIP_NEXT "\xee\x81\x84"
#define ICON_MD_SKIP_PREVIOUS "\xee\x81\x85"
#define ICON_MD_SLIDESHOW "\xee\x90\x9b"
#define ICON_MD_SLOW_MOTION_VIDEO "\xee\x81\xa8"
#define ICON_MD_SMARTPHONE "\xee\x8c\xac"
#define ICON_MD_SMOKE_FREE "\xee\xad\x8a"
#define ICON_MD_SMOKING_ROOMS "\xee\xad\x8b"
#define ICON_MD_SMS "\xee\x98\xa5"
#define ICON_MD_SMS_FAILED "\xee\x98\xa6"
#define ICON_MD_SNOOZE "\xee\x81\x86"
#define ICON_MD_SORT "\xee\x85\xa4"
#define ICON_MD_SORT_BY_ALPHA "\xee\x81\x93"
#define ICON_MD_SPA "\xee\xad\x8c"
#define ICON_MD_SPACE_BAR "\xee\x89\x96"
#define ICON_MD_SPEAKER "\xee\x8c\xad"
#define ICON_MD_SPEAKER_GROUP "\xee\x8c\xae"
#define ICON_MD_SPEAKER_NOTES "\xee\xa3\x8d"
#define ICON_MD_SPEAKER_NOTES_OFF "\xee\xa4\xaa"
#define ICON_MD_SPEAKER_PHONE "\xee\x83\x92"
#define ICON_MD_SPELLCHECK "\xee\xa3\x8e"
#define ICON_MD_STAR "\xee\xa0\xb8"
#define ICON_MD_STAR_BORDER "\xee\xa0\xba"
#define ICON_MD_STAR_HALF "\xee\xa0\xb9"
#define ICON_MD_STARS "\xee\xa3\x90"
#define ICON_MD_STAY_CURRENT_LANDSCAPE "\xee\x83\x93"
#define ICON_MD_STAY_CURRENT_PORTRAIT "\xee\x83\x94"
#define ICON_MD_STAY_PRIMARY_LANDSCAPE "\xee\x83\x95"
#define ICON_MD_STAY_PRIMARY_PORTRAIT "\xee\x83\x96"
#define ICON_MD_STOP "\xee\x81\x87"
#define ICON_MD_STOP_SCREEN_SHARE "\xee\x83\xa3"
#define ICON_MD_STORAGE "\xee\x87\x9b"
#define ICON_MD_STORE "\xee\xa3\x91"
#define ICON_MD_STORE_MALL_DIRECTORY "\xee\x95\xa3"
#define ICON_MD_STRAIGHTEN "\xee\x90\x9c"
#define ICON_MD_STREETVIEW "\xee\x95\xae"
#define ICON_MD_STRIKETHROUGH_S "\xee\x89\x97"
#define ICON_MD_STYLE "\xee\x90\x9d"
#define ICON_MD_SUBDIRECTORY_ARROW_LEFT "\xee\x97\x99"
#define ICON_MD_SUBDIRECTORY_ARROW_RIGHT "\xee\x97\x9a"
#define ICON_MD_SUBJECT "\xee\xa3\x92"
#define ICON_MD_SUBSCRIPTIONS "\xee\x81\xa4"
#define ICON_MD_SUBTITLES "\xee\x81\x88"
#define ICON_MD_SUBWAY "\xee\x95\xaf"
#define ICON_MD_SUPERVISOR_ACCOUNT "\xee\xa3\x93"
#define ICON_MD_SURROUND_SOUND "\xee\x81\x89"
#define ICON_MD_SWAP_CALLS "\xee\x83\x97"
#define ICON_MD_SWAP_HORIZ "\xee\xa3\x94"
#define ICON_MD_SWAP_VERT "\xee\xa3\x95"
#define ICON_MD_SWAP_VERTICAL_CIRCLE "\xee\xa3\x96"
#define ICON_MD_SWITCH_CAMERA "\xee\x90\x9e"
#define ICON_MD_SWITCH_VIDEO "\xee\x90\x9f"
#define ICON_MD_SYNC "\xee\x98\xa7"
#define ICON_MD_SYNC_DISABLED "\xee\x98\xa8"
#define ICON_MD_SYNC_PROBLEM "\xee\x98\xa9"
#define ICON_MD_SYSTEM_UPDATE "\xee\x98\xaa"
#define ICON_MD_SYSTEM_UPDATE_ALT "\xee\xa3\x97"
#define ICON_MD_TAB "\xee\xa3\x98"
#define ICON_MD_TAB_UNSELECTED "\xee\xa3\x99"
#define ICON_MD_TABLET "\xee\x8c\xaf"
#define ICON_MD_TABLET_ANDROID "\xee\x8c\xb0"
#define ICON_MD_TABLET_MAC "\xee\x8c\xb1"
#define ICON_MD_TAG_FACES "\xee\x90\xa0"
#define ICON_MD_TAP_AND_PLAY "\xee\x98\xab"
#define ICON_MD_TERRAIN "\xee\x95\xa4"
#define ICON_MD_TEXT_FIELDS "\xee\x89\xa2"
#define ICON_MD_TEXT_FORMAT "\xee\x85\xa5"
#define ICON_MD_TEXTSMS "\xee\x83\x98"
#define ICON_MD_TEXTURE "\xee\x90\xa1"
#define ICON_MD_THEATERS "\xee\xa3\x9a"
#define ICON_MD_THUMB_DOWN "\xee\xa3\x9b"
#define ICON_MD_THUMB_UP "\xee\xa3\x9c"
#define ICON_MD_THUMBS_UP_DOWN "\xee\xa3\x9d"
#define ICON_MD_TIME_TO_LEAVE "\xee\x98\xac"
#define ICON_MD_TIMELAPSE "\xee\x90\xa2"
#define ICON_MD_TIMELINE "\xee\xa4\xa2"
#define ICON_MD_TIMER "\xee\x90\xa5"
#define ICON_MD_TIMER_10 "\xee\x90\xa3"
#define ICON_MD_TIMER_3 "\xee\x90\xa4"
#define ICON_MD_TIMER_OFF "\xee\x90\xa6"
#define ICON_MD_TITLE "\xee\x89\xa4"
#define ICON_MD_TOC "\xee\xa3\x9e"
#define ICON_MD_TODAY "\xee\xa3\x9f"
#define ICON_MD_TOLL "\xee\xa3\xa0"
#define ICON_MD_TONALITY "\xee\x90\xa7"
#define ICON_MD_TOUCH_APP "\xee\xa4\x93"
#define ICON_MD_TOYS "\xee\x8c\xb2"
#define ICON_MD_TRACK_CHANGES "\xee\xa3\xa1"
#define ICON_MD_TRAFFIC "\xee\x95\xa5"
#define ICON_MD_TRAIN "\xee\x95\xb0"
#define ICON_MD_TRAM "\xee\x95\xb1"
#define ICON_MD_TRANSFER_WITHIN_A_STATION "\xee\x95\xb2"
#define ICON_MD_TRANSFORM "\xee\x90\xa8"
#define ICON_MD_TRANSLATE "\xee\xa3\xa2"
#define ICON_MD_TRENDING_DOWN "\xee\xa3\xa3"
#define ICON_MD_TRENDING_FLAT "\xee\xa3\xa4"
#define ICON_MD_TRENDING_UP "\xee\xa3\xa5"
#define ICON_MD_TUNE "\xee\x90\xa9"
#define ICON_MD_TURNED_IN "\xee\xa3\xa6"
#define ICON_MD_TURNED_IN_NOT "\xee\xa3\xa7"
#define ICON_MD_TV "\xee\x8c\xb3"
#define ICON_MD_UNARCHIVE "\xee\x85\xa9"
#define ICON_MD_UNDO "\xee\x85\xa6"
#define ICON_MD_UNFOLD_LESS "\xee\x97\x96"
#define ICON_MD_UNFOLD_MORE "\xee\x97\x97"
#define ICON_MD_UPDATE "\xee\xa4\xa3"
#define ICON_MD_USB "\xee\x87\xa0"
#define ICON_MD_VERIFIED_USER "\xee\xa3\xa8"
#define ICON_MD_VERTICAL_ALIGN_BOTTOM "\xee\x89\x98"
#define ICON_MD_VERTICAL_ALIGN_CENTER "\xee\x89\x99"
#define ICON_MD_VERTICAL_ALIGN_TOP "\xee\x89\x9a"
#define ICON_MD_VIBRATION "\xee\x98\xad"
#define ICON_MD_VIDEO_CALL "\xee\x81\xb0"
#define ICON_MD_VIDEO_LABEL "\xee\x81\xb1"
#define ICON_MD_VIDEO_LIBRARY "\xee\x81\x8a"
#define ICON_MD_VIDEOCAM "\xee\x81\x8b"
#define ICON_MD_VIDEOCAM_OFF "\xee\x81\x8c"
#define ICON_MD_VIDEOGAME_ASSET "\xee\x8c\xb8"
#define ICON_MD_VIEW_AGENDA "\xee\xa3\xa9"
#define ICON_MD_VIEW_ARRAY "\xee\xa3\xaa"
#define ICON_MD_VIEW_CAROUSEL "\xee\xa3\xab"
#define ICON_MD_VIEW_COLUMN "\xee\xa3\xac"
#define ICON_MD_VIEW_COMFY "\xee\x90\xaa"
#define ICON_MD_VIEW_COMPACT "\xee\x90\xab"
#define ICON_MD_VIEW_DAY "\xee\xa3\xad"
#define ICON_MD_VIEW_HEADLINE "\xee\xa3\xae"
#define ICON_MD_VIEW_LIST "\xee\xa3\xaf"
#define ICON_MD_VIEW_MODULE "\xee\xa3\xb0"
#define ICON_MD_VIEW_QUILT "\xee\xa3\xb1"
#define ICON_MD_VIEW_STREAM "\xee\xa3\xb2"
#define ICON_MD_VIEW_WEEK "\xee\xa3\xb3"
#define ICON_MD_VIGNETTE "\xee\x90\xb5"
#define ICON_MD_VISIBILITY "\xee\xa3\xb4"
#define ICON_MD_VISIBILITY_OFF "\xee\xa3\xb5"
#define ICON_MD_VOICE_CHAT "\xee\x98\xae"
#define ICON_MD_VOICEMAIL "\xee\x83\x99"
#define ICON_MD_VOLUME_DOWN "\xee\x81\x8d"
#define ICON_MD_VOLUME_MUTE "\xee\x81\x8e"
#define ICON_MD_VOLUME_OFF "\xee\x81\x8f"
#define ICON_MD_VOLUME_UP "\xee\x81\x90"
#define ICON_MD_VPN_KEY "\xee\x83\x9a"
#define ICON_MD_VPN_LOCK "\xee\x98\xaf"
#define ICON_MD_WALLPAPER "\xee\x86\xbc"
#define ICON_MD_WARNING "\xee\x80\x82"
#define ICON_MD_WATCH "\xee\x8c\xb4"
#define ICON_MD_WATCH_LATER "\xee\xa4\xa4"
#define ICON_MD_WB_AUTO "\xee\x90\xac"
#define ICON_MD_WB_CLOUDY "\xee\x90\xad"
#define ICON_MD_WB_INCANDESCENT "\xee\x90\xae"
#define ICON_MD_WB_IRIDESCENT "\xee\x90\xb6"
#define ICON_MD_WB_SUNNY "\xee\x90\xb0"
#define ICON_MD_WC "\xee\x98\xbd"
#define ICON_MD_WEB "\xee\x81\x91"
#define ICON_MD_WEB_ASSET "\xee\x81\xa9"
#define ICON_MD_WEEKEND "\xee\x85\xab"
#define ICON_MD_WHATSHOT "\xee\xa0\x8e"
#define ICON_MD_WIDGETS "\xee\x86\xbd"
#define ICON_MD_WIFI "\xee\x98\xbe"
#define ICON_MD_WIFI_LOCK "\xee\x87\xa1"
#define ICON_MD_WIFI_TETHERING "\xee\x87\xa2"
#define ICON_MD_WORK "\xee\xa3\xb9"
#define ICON_MD_WRAP_TEXT "\xee\x89\x9b"
#define ICON_MD_YOUTUBE_SEARCHED_FOR "\xee\xa3\xba"
#define ICON_MD_ZOOM_IN "\xee\xa3\xbf"
#define ICON_MD_ZOOM_OUT "\xee\xa4\x80"
#define ICON_MD_ZOOM_OUT_MAP "\xee\x95\xab"
