// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C#
// from https://raw.githubusercontent.com/ForkAwesome/Fork-Awesome/master/src/icons/icons.yml
// for use with https://github.com/ForkAwesome/Fork-Awesome/blob/master/fonts/forkawesome-webfont.ttf
namespace IconFonts
{
    public class ForkAwesome
    {
        public const string FontIconFileName = "forkawesome-webfont.ttf";

        public const int IconMin = 0xf000;
        public const int IconMax = 0xf35f;
        public const string Glass = "\uf000";
        public const string Music = "\uf001";
        public const string Search = "\uf002";
        public const string EnvelopeO = "\uf003";
        public const string Heart = "\uf004";
        public const string Star = "\uf005";
        public const string StarO = "\uf006";
        public const string User = "\uf007";
        public const string Film = "\uf008";
        public const string ThLarge = "\uf009";
        public const string Th = "\uf00a";
        public const string ThList = "\uf00b";
        public const string Check = "\uf00c";
        public const string Times = "\uf00d";
        public const string SearchPlus = "\uf00e";
        public const string SearchMinus = "\uf010";
        public const string PowerOff = "\uf011";
        public const string Signal = "\uf012";
        public const string Cog = "\uf013";
        public const string TrashO = "\uf014";
        public const string Home = "\uf015";
        public const string FileO = "\uf016";
        public const string ClockO = "\uf017";
        public const string Road = "\uf018";
        public const string Download = "\uf019";
        public const string ArrowCircleODown = "\uf01a";
        public const string ArrowCircleOUp = "\uf01b";
        public const string Inbox = "\uf01c";
        public const string PlayCircleO = "\uf01d";
        public const string Repeat = "\uf01e";
        public const string Refresh = "\uf021";
        public const string ListAlt = "\uf022";
        public const string Lock = "\uf023";
        public const string Flag = "\uf024";
        public const string Headphones = "\uf025";
        public const string VolumeOff = "\uf026";
        public const string VolumeDown = "\uf027";
        public const string VolumeUp = "\uf028";
        public const string Qrcode = "\uf029";
        public const string Barcode = "\uf02a";
        public const string Tag = "\uf02b";
        public const string Tags = "\uf02c";
        public const string Book = "\uf02d";
        public const string Bookmark = "\uf02e";
        public const string Print = "\uf02f";
        public const string Camera = "\uf030";
        public const string Font = "\uf031";
        public const string Bold = "\uf032";
        public const string Italic = "\uf033";
        public const string TextHeight = "\uf034";
        public const string TextWidth = "\uf035";
        public const string AlignLeft = "\uf036";
        public const string AlignCenter = "\uf037";
        public const string AlignRight = "\uf038";
        public const string AlignJustify = "\uf039";
        public const string List = "\uf03a";
        public const string Outdent = "\uf03b";
        public const string Indent = "\uf03c";
        public const string VideoCamera = "\uf03d";
        public const string PictureO = "\uf03e";
        public const string Pencil = "\uf040";
        public const string MapMarker = "\uf041";
        public const string Adjust = "\uf042";
        public const string Tint = "\uf043";
        public const string PencilSquareO = "\uf044";
        public const string ShareSquareO = "\uf045";
        public const string CheckSquareO = "\uf046";
        public const string Arrows = "\uf047";
        public const string StepBackward = "\uf048";
        public const string FastBackward = "\uf049";
        public const string Backward = "\uf04a";
        public const string Play = "\uf04b";
        public const string Pause = "\uf04c";
        public const string Stop = "\uf04d";
        public const string Forward = "\uf04e";
        public const string FastForward = "\uf050";
        public const string StepForward = "\uf051";
        public const string Eject = "\uf052";
        public const string ChevronLeft = "\uf053";
        public const string ChevronRight = "\uf054";
        public const string PlusCircle = "\uf055";
        public const string MinusCircle = "\uf056";
        public const string TimesCircle = "\uf057";
        public const string CheckCircle = "\uf058";
        public const string QuestionCircle = "\uf059";
        public const string InfoCircle = "\uf05a";
        public const string Crosshairs = "\uf05b";
        public const string TimesCircleO = "\uf05c";
        public const string CheckCircleO = "\uf05d";
        public const string Ban = "\uf05e";
        public const string ArrowLeft = "\uf060";
        public const string ArrowRight = "\uf061";
        public const string ArrowUp = "\uf062";
        public const string ArrowDown = "\uf063";
        public const string Share = "\uf064";
        public const string Expand = "\uf065";
        public const string Compress = "\uf066";
        public const string Plus = "\uf067";
        public const string Minus = "\uf068";
        public const string Asterisk = "\uf069";
        public const string ExclamationCircle = "\uf06a";
        public const string Gift = "\uf06b";
        public const string Leaf = "\uf06c";
        public const string Fire = "\uf06d";
        public const string Eye = "\uf06e";
        public const string EyeSlash = "\uf070";
        public const string ExclamationTriangle = "\uf071";
        public const string Plane = "\uf072";
        public const string Calendar = "\uf073";
        public const string Random = "\uf074";
        public const string Comment = "\uf075";
        public const string Magnet = "\uf076";
        public const string ChevronUp = "\uf077";
        public const string ChevronDown = "\uf078";
        public const string Retweet = "\uf079";
        public const string ShoppingCart = "\uf07a";
        public const string Folder = "\uf07b";
        public const string FolderOpen = "\uf07c";
        public const string ArrowsV = "\uf07d";
        public const string ArrowsH = "\uf07e";
        public const string BarChart = "\uf080";
        public const string TwitterSquare = "\uf081";
        public const string FacebookSquare = "\uf082";
        public const string CameraRetro = "\uf083";
        public const string Key = "\uf084";
        public const string Cogs = "\uf085";
        public const string Comments = "\uf086";
        public const string ThumbsOUp = "\uf087";
        public const string ThumbsODown = "\uf088";
        public const string StarHalf = "\uf089";
        public const string HeartO = "\uf08a";
        public const string SignOut = "\uf08b";
        public const string LinkedinSquare = "\uf08c";
        public const string ThumbTack = "\uf08d";
        public const string ExternalLink = "\uf08e";
        public const string SignIn = "\uf090";
        public const string Trophy = "\uf091";
        public const string GithubSquare = "\uf092";
        public const string Upload = "\uf093";
        public const string LemonO = "\uf094";
        public const string Phone = "\uf095";
        public const string SquareO = "\uf096";
        public const string BookmarkO = "\uf097";
        public const string PhoneSquare = "\uf098";
        public const string Twitter = "\uf099";
        public const string Facebook = "\uf09a";
        public const string Github = "\uf09b";
        public const string Unlock = "\uf09c";
        public const string CreditCard = "\uf09d";
        public const string Rss = "\uf09e";
        public const string HddO = "\uf0a0";
        public const string Bullhorn = "\uf0a1";
        public const string BellO = "\uf0f3";
        public const string Certificate = "\uf0a3";
        public const string HandORight = "\uf0a4";
        public const string HandOLeft = "\uf0a5";
        public const string HandOUp = "\uf0a6";
        public const string HandODown = "\uf0a7";
        public const string ArrowCircleLeft = "\uf0a8";
        public const string ArrowCircleRight = "\uf0a9";
        public const string ArrowCircleUp = "\uf0aa";
        public const string ArrowCircleDown = "\uf0ab";
        public const string Globe = "\uf0ac";
        public const string GlobeE = "\uf304";
        public const string GlobeW = "\uf305";
        public const string Wrench = "\uf0ad";
        public const string Tasks = "\uf0ae";
        public const string Filter = "\uf0b0";
        public const string Briefcase = "\uf0b1";
        public const string ArrowsAlt = "\uf0b2";
        public const string Users = "\uf0c0";
        public const string Link = "\uf0c1";
        public const string Cloud = "\uf0c2";
        public const string Flask = "\uf0c3";
        public const string Scissors = "\uf0c4";
        public const string FilesO = "\uf0c5";
        public const string Paperclip = "\uf0c6";
        public const string FloppyO = "\uf0c7";
        public const string Square = "\uf0c8";
        public const string Bars = "\uf0c9";
        public const string ListUl = "\uf0ca";
        public const string ListOl = "\uf0cb";
        public const string Strikethrough = "\uf0cc";
        public const string Underline = "\uf0cd";
        public const string Table = "\uf0ce";
        public const string Magic = "\uf0d0";
        public const string Truck = "\uf0d1";
        public const string Pinterest = "\uf0d2";
        public const string PinterestSquare = "\uf0d3";
        public const string GooglePlusSquare = "\uf0d4";
        public const string GooglePlus = "\uf0d5";
        public const string Money = "\uf0d6";
        public const string CaretDown = "\uf0d7";
        public const string CaretUp = "\uf0d8";
        public const string CaretLeft = "\uf0d9";
        public const string CaretRight = "\uf0da";
        public const string Columns = "\uf0db";
        public const string Sort = "\uf0dc";
        public const string SortDesc = "\uf0dd";
        public const string SortAsc = "\uf0de";
        public const string Envelope = "\uf0e0";
        public const string Linkedin = "\uf0e1";
        public const string Undo = "\uf0e2";
        public const string Gavel = "\uf0e3";
        public const string Tachometer = "\uf0e4";
        public const string CommentO = "\uf0e5";
        public const string CommentsO = "\uf0e6";
        public const string Bolt = "\uf0e7";
        public const string Sitemap = "\uf0e8";
        public const string Umbrella = "\uf0e9";
        public const string Clipboard = "\uf0ea";
        public const string LightbulbO = "\uf0eb";
        public const string Exchange = "\uf0ec";
        public const string CloudDownload = "\uf0ed";
        public const string CloudUpload = "\uf0ee";
        public const string UserMd = "\uf0f0";
        public const string Stethoscope = "\uf0f1";
        public const string Suitcase = "\uf0f2";
        public const string Bell = "\uf0a2";
        public const string Coffee = "\uf0f4";
        public const string Cutlery = "\uf0f5";
        public const string FileTextO = "\uf0f6";
        public const string BuildingO = "\uf0f7";
        public const string HospitalO = "\uf0f8";
        public const string Ambulance = "\uf0f9";
        public const string Medkit = "\uf0fa";
        public const string FighterJet = "\uf0fb";
        public const string Beer = "\uf0fc";
        public const string HSquare = "\uf0fd";
        public const string PlusSquare = "\uf0fe";
        public const string AngleDoubleLeft = "\uf100";
        public const string AngleDoubleRight = "\uf101";
        public const string AngleDoubleUp = "\uf102";
        public const string AngleDoubleDown = "\uf103";
        public const string AngleLeft = "\uf104";
        public const string AngleRight = "\uf105";
        public const string AngleUp = "\uf106";
        public const string AngleDown = "\uf107";
        public const string Desktop = "\uf108";
        public const string Laptop = "\uf109";
        public const string Tablet = "\uf10a";
        public const string Mobile = "\uf10b";
        public const string CircleO = "\uf10c";
        public const string QuoteLeft = "\uf10d";
        public const string QuoteRight = "\uf10e";
        public const string Spinner = "\uf110";
        public const string Circle = "\uf111";
        public const string Reply = "\uf112";
        public const string GithubAlt = "\uf113";
        public const string FolderO = "\uf114";
        public const string FolderOpenO = "\uf115";
        public const string SmileO = "\uf118";
        public const string FrownO = "\uf119";
        public const string MehO = "\uf11a";
        public const string Gamepad = "\uf11b";
        public const string KeyboardO = "\uf11c";
        public const string FlagO = "\uf11d";
        public const string FlagCheckered = "\uf11e";
        public const string Terminal = "\uf120";
        public const string Code = "\uf121";
        public const string ReplyAll = "\uf122";
        public const string StarHalfO = "\uf123";
        public const string LocationArrow = "\uf124";
        public const string Crop = "\uf125";
        public const string CodeFork = "\uf126";
        public const string ChainBroken = "\uf127";
        public const string Question = "\uf128";
        public const string Info = "\uf129";
        public const string Exclamation = "\uf12a";
        public const string Superscript = "\uf12b";
        public const string Subscript = "\uf12c";
        public const string Eraser = "\uf12d";
        public const string PuzzlePiece = "\uf12e";
        public const string Microphone = "\uf130";
        public const string MicrophoneSlash = "\uf131";
        public const string Shield = "\uf132";
        public const string CalendarO = "\uf133";
        public const string FireExtinguisher = "\uf134";
        public const string Rocket = "\uf135";
        public const string Maxcdn = "\uf136";
        public const string ChevronCircleLeft = "\uf137";
        public const string ChevronCircleRight = "\uf138";
        public const string ChevronCircleUp = "\uf139";
        public const string ChevronCircleDown = "\uf13a";
        public const string Html5 = "\uf13b";
        public const string Css3 = "\uf13c";
        public const string Anchor = "\uf13d";
        public const string UnlockAlt = "\uf13e";
        public const string Bullseye = "\uf140";
        public const string EllipsisH = "\uf141";
        public const string EllipsisV = "\uf142";
        public const string RssSquare = "\uf143";
        public const string PlayCircle = "\uf144";
        public const string Ticket = "\uf145";
        public const string MinusSquare = "\uf146";
        public const string MinusSquareO = "\uf147";
        public const string LevelUp = "\uf148";
        public const string LevelDown = "\uf149";
        public const string CheckSquare = "\uf14a";
        public const string PencilSquare = "\uf14b";
        public const string ExternalLinkSquare = "\uf14c";
        public const string ShareSquare = "\uf14d";
        public const string Compass = "\uf14e";
        public const string CaretSquareODown = "\uf150";
        public const string CaretSquareOUp = "\uf151";
        public const string CaretSquareORight = "\uf152";
        public const string Eur = "\uf153";
        public const string Gbp = "\uf154";
        public const string Usd = "\uf155";
        public const string Inr = "\uf156";
        public const string Jpy = "\uf157";
        public const string Rub = "\uf158";
        public const string Krw = "\uf159";
        public const string Btc = "\uf15a";
        public const string File = "\uf15b";
        public const string FileText = "\uf15c";
        public const string SortAlphaAsc = "\uf15d";
        public const string SortAlphaDesc = "\uf15e";
        public const string SortAmountAsc = "\uf160";
        public const string SortAmountDesc = "\uf161";
        public const string SortNumericAsc = "\uf162";
        public const string SortNumericDesc = "\uf163";
        public const string ThumbsUp = "\uf164";
        public const string ThumbsDown = "\uf165";
        public const string YoutubeSquare = "\uf166";
        public const string Youtube = "\uf167";
        public const string Xing = "\uf168";
        public const string XingSquare = "\uf169";
        public const string YoutubePlay = "\uf16a";
        public const string Dropbox = "\uf16b";
        public const string StackOverflow = "\uf16c";
        public const string Instagram = "\uf16d";
        public const string Flickr = "\uf16e";
        public const string Adn = "\uf170";
        public const string Bitbucket = "\uf171";
        public const string BitbucketSquare = "\uf172";
        public const string Tumblr = "\uf173";
        public const string TumblrSquare = "\uf174";
        public const string LongArrowDown = "\uf175";
        public const string LongArrowUp = "\uf176";
        public const string LongArrowLeft = "\uf177";
        public const string LongArrowRight = "\uf178";
        public const string Apple = "\uf179";
        public const string Windows = "\uf17a";
        public const string Android = "\uf17b";
        public const string Linux = "\uf17c";
        public const string Dribbble = "\uf17d";
        public const string Skype = "\uf17e";
        public const string Foursquare = "\uf180";
        public const string Trello = "\uf181";
        public const string Female = "\uf182";
        public const string Male = "\uf183";
        public const string Gratipay = "\uf184";
        public const string SunO = "\uf185";
        public const string MoonO = "\uf186";
        public const string Archive = "\uf187";
        public const string Bug = "\uf188";
        public const string Vk = "\uf189";
        public const string Weibo = "\uf18a";
        public const string Renren = "\uf18b";
        public const string Pagelines = "\uf18c";
        public const string StackExchange = "\uf18d";
        public const string ArrowCircleORight = "\uf18e";
        public const string ArrowCircleOLeft = "\uf190";
        public const string CaretSquareOLeft = "\uf191";
        public const string DotCircleO = "\uf192";
        public const string Wheelchair = "\uf193";
        public const string VimeoSquare = "\uf194";
        public const string Try = "\uf195";
        public const string PlusSquareO = "\uf196";
        public const string SpaceShuttle = "\uf197";
        public const string Slack = "\uf198";
        public const string EnvelopeSquare = "\uf199";
        public const string Wordpress = "\uf19a";
        public const string Openid = "\uf19b";
        public const string University = "\uf19c";
        public const string GraduationCap = "\uf19d";
        public const string Yahoo = "\uf19e";
        public const string Google = "\uf1a0";
        public const string Reddit = "\uf1a1";
        public const string RedditSquare = "\uf1a2";
        public const string StumbleuponCircle = "\uf1a3";
        public const string Stumbleupon = "\uf1a4";
        public const string Delicious = "\uf1a5";
        public const string Digg = "\uf1a6";
        public const string Drupal = "\uf1a9";
        public const string Joomla = "\uf1aa";
        public const string Language = "\uf1ab";
        public const string Fax = "\uf1ac";
        public const string Building = "\uf1ad";
        public const string Child = "\uf1ae";
        public const string Paw = "\uf1b0";
        public const string Spoon = "\uf1b1";
        public const string Cube = "\uf1b2";
        public const string Cubes = "\uf1b3";
        public const string Behance = "\uf1b4";
        public const string BehanceSquare = "\uf1b5";
        public const string Steam = "\uf1b6";
        public const string SteamSquare = "\uf1b7";
        public const string Recycle = "\uf1b8";
        public const string Car = "\uf1b9";
        public const string Taxi = "\uf1ba";
        public const string Tree = "\uf1bb";
        public const string Spotify = "\uf1bc";
        public const string Deviantart = "\uf1bd";
        public const string Soundcloud = "\uf1be";
        public const string Database = "\uf1c0";
        public const string FilePdfO = "\uf1c1";
        public const string FileWordO = "\uf1c2";
        public const string FileExcelO = "\uf1c3";
        public const string FilePowerpointO = "\uf1c4";
        public const string FileImageO = "\uf1c5";
        public const string FileArchiveO = "\uf1c6";
        public const string FileAudioO = "\uf1c7";
        public const string FileVideoO = "\uf1c8";
        public const string FileCodeO = "\uf1c9";
        public const string Vine = "\uf1ca";
        public const string Codepen = "\uf1cb";
        public const string Jsfiddle = "\uf1cc";
        public const string LifeRing = "\uf1cd";
        public const string CircleONotch = "\uf1ce";
        public const string Rebel = "\uf1d0";
        public const string Empire = "\uf1d1";
        public const string GitSquare = "\uf1d2";
        public const string Git = "\uf1d3";
        public const string HackerNews = "\uf1d4";
        public const string TencentWeibo = "\uf1d5";
        public const string Qq = "\uf1d6";
        public const string Weixin = "\uf1d7";
        public const string PaperPlane = "\uf1d8";
        public const string PaperPlaneO = "\uf1d9";
        public const string History = "\uf1da";
        public const string CircleThin = "\uf1db";
        public const string Header = "\uf1dc";
        public const string Paragraph = "\uf1dd";
        public const string Sliders = "\uf1de";
        public const string ShareAlt = "\uf1e0";
        public const string ShareAltSquare = "\uf1e1";
        public const string Bomb = "\uf1e2";
        public const string FutbolO = "\uf1e3";
        public const string Tty = "\uf1e4";
        public const string Binoculars = "\uf1e5";
        public const string Plug = "\uf1e6";
        public const string Slideshare = "\uf1e7";
        public const string Twitch = "\uf1e8";
        public const string Yelp = "\uf1e9";
        public const string NewspaperO = "\uf1ea";
        public const string Wifi = "\uf1eb";
        public const string Calculator = "\uf1ec";
        public const string Paypal = "\uf1ed";
        public const string GoogleWallet = "\uf1ee";
        public const string CcVisa = "\uf1f0";
        public const string CcMastercard = "\uf1f1";
        public const string CcDiscover = "\uf1f2";
        public const string CcAmex = "\uf1f3";
        public const string CcPaypal = "\uf1f4";
        public const string CcStripe = "\uf1f5";
        public const string BellSlash = "\uf1f6";
        public const string BellSlashO = "\uf1f7";
        public const string Trash = "\uf1f8";
        public const string Copyright = "\uf1f9";
        public const string At = "\uf1fa";
        public const string Eyedropper = "\uf1fb";
        public const string PaintBrush = "\uf1fc";
        public const string BirthdayCake = "\uf1fd";
        public const string AreaChart = "\uf1fe";
        public const string PieChart = "\uf200";
        public const string LineChart = "\uf201";
        public const string Lastfm = "\uf202";
        public const string LastfmSquare = "\uf203";
        public const string ToggleOff = "\uf204";
        public const string ToggleOn = "\uf205";
        public const string Bicycle = "\uf206";
        public const string Bus = "\uf207";
        public const string Ioxhost = "\uf208";
        public const string Angellist = "\uf209";
        public const string Cc = "\uf20a";
        public const string Ils = "\uf20b";
        public const string Meanpath = "\uf20c";
        public const string Buysellads = "\uf20d";
        public const string Connectdevelop = "\uf20e";
        public const string Dashcube = "\uf210";
        public const string Forumbee = "\uf211";
        public const string Leanpub = "\uf212";
        public const string Sellsy = "\uf213";
        public const string Shirtsinbulk = "\uf214";
        public const string Simplybuilt = "\uf215";
        public const string Skyatlas = "\uf216";
        public const string CartPlus = "\uf217";
        public const string CartArrowDown = "\uf218";
        public const string Diamond = "\uf219";
        public const string Ship = "\uf21a";
        public const string UserSecret = "\uf21b";
        public const string Motorcycle = "\uf21c";
        public const string StreetView = "\uf21d";
        public const string Heartbeat = "\uf21e";
        public const string Venus = "\uf221";
        public const string Mars = "\uf222";
        public const string Mercury = "\uf223";
        public const string Transgender = "\uf224";
        public const string TransgenderAlt = "\uf225";
        public const string VenusDouble = "\uf226";
        public const string MarsDouble = "\uf227";
        public const string VenusMars = "\uf228";
        public const string MarsStroke = "\uf229";
        public const string MarsStrokeV = "\uf22a";
        public const string MarsStrokeH = "\uf22b";
        public const string Neuter = "\uf22c";
        public const string Genderless = "\uf22d";
        public const string FacebookOfficial = "\uf230";
        public const string PinterestP = "\uf231";
        public const string Whatsapp = "\uf232";
        public const string Server = "\uf233";
        public const string UserPlus = "\uf234";
        public const string UserTimes = "\uf235";
        public const string Bed = "\uf236";
        public const string Viacoin = "\uf237";
        public const string Train = "\uf238";
        public const string Subway = "\uf239";
        public const string Medium = "\uf23a";
        public const string MediumSquare = "\uf2f8";
        public const string YCombinator = "\uf23b";
        public const string OptinMonster = "\uf23c";
        public const string Opencart = "\uf23d";
        public const string Expeditedssl = "\uf23e";
        public const string BatteryFull = "\uf240";
        public const string BatteryThreeQuarters = "\uf241";
        public const string BatteryHalf = "\uf242";
        public const string BatteryQuarter = "\uf243";
        public const string BatteryEmpty = "\uf244";
        public const string MousePointer = "\uf245";
        public const string ICursor = "\uf246";
        public const string ObjectGroup = "\uf247";
        public const string ObjectUngroup = "\uf248";
        public const string StickyNote = "\uf249";
        public const string StickyNoteO = "\uf24a";
        public const string CcJcb = "\uf24b";
        public const string CcDinersClub = "\uf24c";
        public const string Clone = "\uf24d";
        public const string BalanceScale = "\uf24e";
        public const string HourglassO = "\uf250";
        public const string HourglassStart = "\uf251";
        public const string HourglassHalf = "\uf252";
        public const string HourglassEnd = "\uf253";
        public const string Hourglass = "\uf254";
        public const string HandRockO = "\uf255";
        public const string HandPaperO = "\uf256";
        public const string HandScissorsO = "\uf257";
        public const string HandLizardO = "\uf258";
        public const string HandSpockO = "\uf259";
        public const string HandPointerO = "\uf25a";
        public const string HandPeaceO = "\uf25b";
        public const string Trademark = "\uf25c";
        public const string Registered = "\uf25d";
        public const string CreativeCommons = "\uf25e";
        public const string Gg = "\uf260";
        public const string GgCircle = "\uf261";
        public const string Tripadvisor = "\uf262";
        public const string Odnoklassniki = "\uf263";
        public const string OdnoklassnikiSquare = "\uf264";
        public const string GetPocket = "\uf265";
        public const string WikipediaW = "\uf266";
        public const string Safari = "\uf267";
        public const string Chrome = "\uf268";
        public const string Firefox = "\uf269";
        public const string Opera = "\uf26a";
        public const string InternetExplorer = "\uf26b";
        public const string Television = "\uf26c";
        public const string Contao = "\uf26d";
        public const string The500px = "\uf26e";
        public const string Amazon = "\uf270";
        public const string CalendarPlusO = "\uf271";
        public const string CalendarMinusO = "\uf272";
        public const string CalendarTimesO = "\uf273";
        public const string CalendarCheckO = "\uf274";
        public const string Industry = "\uf275";
        public const string MapPin = "\uf276";
        public const string MapSigns = "\uf277";
        public const string MapO = "\uf278";
        public const string Map = "\uf279";
        public const string Commenting = "\uf27a";
        public const string CommentingO = "\uf27b";
        public const string Houzz = "\uf27c";
        public const string Vimeo = "\uf27d";
        public const string BlackTie = "\uf27e";
        public const string Fonticons = "\uf280";
        public const string RedditAlien = "\uf281";
        public const string Edge = "\uf282";
        public const string CreditCardAlt = "\uf283";
        public const string Codiepie = "\uf284";
        public const string Modx = "\uf285";
        public const string FortAwesome = "\uf286";
        public const string Usb = "\uf287";
        public const string ProductHunt = "\uf288";
        public const string Mixcloud = "\uf289";
        public const string Scribd = "\uf28a";
        public const string PauseCircle = "\uf28b";
        public const string PauseCircleO = "\uf28c";
        public const string StopCircle = "\uf28d";
        public const string StopCircleO = "\uf28e";
        public const string ShoppingBag = "\uf290";
        public const string ShoppingBasket = "\uf291";
        public const string Hashtag = "\uf292";
        public const string Bluetooth = "\uf293";
        public const string BluetoothB = "\uf294";
        public const string Percent = "\uf295";
        public const string Gitlab = "\uf296";
        public const string Wpbeginner = "\uf297";
        public const string Wpforms = "\uf298";
        public const string Envira = "\uf299";
        public const string UniversalAccess = "\uf29a";
        public const string WheelchairAlt = "\uf29b";
        public const string QuestionCircleO = "\uf29c";
        public const string Blind = "\uf29d";
        public const string AudioDescription = "\uf29e";
        public const string VolumeControlPhone = "\uf2a0";
        public const string Braille = "\uf2a1";
        public const string AssistiveListeningSystems = "\uf2a2";
        public const string AmericanSignLanguageInterpreting = "\uf2a3";
        public const string Deaf = "\uf2a4";
        public const string Glide = "\uf2a5";
        public const string GlideG = "\uf2a6";
        public const string SignLanguage = "\uf2a7";
        public const string LowVision = "\uf2a8";
        public const string Viadeo = "\uf2a9";
        public const string ViadeoSquare = "\uf2aa";
        public const string Snapchat = "\uf2ab";
        public const string SnapchatGhost = "\uf2ac";
        public const string SnapchatSquare = "\uf2ad";
        public const string FirstOrder = "\uf2b0";
        public const string Yoast = "\uf2b1";
        public const string Themeisle = "\uf2b2";
        public const string GooglePlusOfficial = "\uf2b3";
        public const string FontAwesome = "\uf2b4";
        public const string HandshakeO = "\uf2b5";
        public const string EnvelopeOpen = "\uf2b6";
        public const string EnvelopeOpenO = "\uf2b7";
        public const string Linode = "\uf2b8";
        public const string AddressBook = "\uf2b9";
        public const string AddressBookO = "\uf2ba";
        public const string AddressCard = "\uf2bb";
        public const string AddressCardO = "\uf2bc";
        public const string UserCircle = "\uf2bd";
        public const string UserCircleO = "\uf2be";
        public const string UserO = "\uf2c0";
        public const string IdBadge = "\uf2c1";
        public const string IdCard = "\uf2c2";
        public const string IdCardO = "\uf2c3";
        public const string Quora = "\uf2c4";
        public const string FreeCodeCamp = "\uf2c5";
        public const string Telegram = "\uf2c6";
        public const string ThermometerFull = "\uf2c7";
        public const string ThermometerThreeQuarters = "\uf2c8";
        public const string ThermometerHalf = "\uf2c9";
        public const string ThermometerQuarter = "\uf2ca";
        public const string ThermometerEmpty = "\uf2cb";
        public const string Shower = "\uf2cc";
        public const string Bath = "\uf2cd";
        public const string Podcast = "\uf2ce";
        public const string WindowMaximize = "\uf2d0";
        public const string WindowMinimize = "\uf2d1";
        public const string WindowRestore = "\uf2d2";
        public const string WindowClose = "\uf2d3";
        public const string WindowCloseO = "\uf2d4";
        public const string Bandcamp = "\uf2d5";
        public const string Grav = "\uf2d6";
        public const string Etsy = "\uf2d7";
        public const string Imdb = "\uf2d8";
        public const string Ravelry = "\uf2d9";
        public const string Eercast = "\uf2da";
        public const string Microchip = "\uf2db";
        public const string SnowflakeO = "\uf2dc";
        public const string Superpowers = "\uf2dd";
        public const string Wpexplorer = "\uf2de";
        public const string Meetup = "\uf2e0";
        public const string Mastodon = "\uf2e1";
        public const string MastodonAlt = "\uf2e2";
        public const string ForkAwesomeIcon = "\uf2e3";
        public const string Peertube = "\uf2e4";
        public const string Diaspora = "\uf2e5";
        public const string Friendica = "\uf2e6";
        public const string GnuSocial = "\uf2e7";
        public const string LiberapaySquare = "\uf2e8";
        public const string Liberapay = "\uf2e9";
        public const string Scuttlebutt = "\uf2ea";
        public const string Hubzilla = "\uf2eb";
        public const string SocialHome = "\uf2ec";
        public const string Artstation = "\uf2ed";
        public const string Discord = "\uf2ee";
        public const string DiscordAlt = "\uf2ef";
        public const string Patreon = "\uf2f0";
        public const string Snowdrift = "\uf2f1";
        public const string Activitypub = "\uf2f2";
        public const string Ethereum = "\uf2f3";
        public const string Keybase = "\uf2f4";
        public const string Shaarli = "\uf2f5";
        public const string ShaarliO = "\uf2f6";
        public const string KeyModern = "\uf2f7";
        public const string Xmpp = "\uf2f9";
        public const string ArchiveOrg = "\uf2fc";
        public const string Freedombox = "\uf2fd";
        public const string FacebookMessenger = "\uf2fe";
        public const string Debian = "\uf2ff";
        public const string MastodonSquare = "\uf300";
        public const string Tipeee = "\uf301";
        public const string React = "\uf302";
        public const string Dogmazic = "\uf303";
        public const string Zotero = "\uf309";
        public const string Nodejs = "\uf308";
        public const string Nextcloud = "\uf306";
        public const string NextcloudSquare = "\uf307";
        public const string Hackaday = "\uf30a";
        public const string Laravel = "\uf30b";
        public const string Signalapp = "\uf30c";
        public const string Gnupg = "\uf30d";
        public const string Php = "\uf30e";
        public const string Ffmpeg = "\uf30f";
        public const string Joplin = "\uf310";
        public const string Syncthing = "\uf311";
        public const string Inkscape = "\uf312";
        public const string MatrixOrg = "\uf313";
        public const string Pixelfed = "\uf314";
        public const string Bootstrap = "\uf315";
        public const string DevTo = "\uf316";
        public const string Hashnode = "\uf317";
        public const string Jirafeau = "\uf318";
        public const string Emby = "\uf319";
        public const string Wikidata = "\uf31a";
        public const string Gimp = "\uf31b";
        public const string C = "\uf31c";
        public const string Digitalocean = "\uf31d";
        public const string Att = "\uf31e";
        public const string Gitea = "\uf31f";
        public const string FileEpub = "\uf321";
        public const string Python = "\uf322";
        public const string Archlinux = "\uf323";
        public const string Pleroma = "\uf324";
        public const string Unsplash = "\uf325";
        public const string Hackster = "\uf326";
        public const string SpellCheck = "\uf327";
        public const string Moon = "\uf328";
        public const string Sun = "\uf329";
        public const string FDroid = "\uf32a";
        public const string Biometric = "\uf32b";
        public const string Wire = "\uf32c";
        public const string TorOnion = "\uf32e";
        public const string VolumeMute = "\uf32f";
        public const string BellRinging = "\uf32d";
        public const string BellRingingO = "\uf330";
        public const string Hal = "\uf333";
        public const string Jupyter = "\uf335";
        public const string Julia = "\uf334";
        public const string Classicpress = "\uf331";
        public const string ClassicpressCircle = "\uf332";
        public const string OpenCollective = "\uf336";
        public const string Orcid = "\uf337";
        public const string Researchgate = "\uf338";
        public const string Funkwhale = "\uf339";
        public const string Askfm = "\uf33a";
        public const string Blockstack = "\uf33b";
        public const string Boardgamegeek = "\uf33c";
        public const string Bunny = "\uf35f";
        public const string Buymeacoffee = "\uf33d";
        public const string CcBy = "\uf33e";
        public const string CcCc = "\uf33f";
        public const string CcNcEu = "\uf341";
        public const string CcNcJp = "\uf342";
        public const string CcNc = "\uf340";
        public const string CcNd = "\uf343";
        public const string CcPd = "\uf344";
        public const string CcRemix = "\uf345";
        public const string CcSa = "\uf346";
        public const string CcShare = "\uf347";
        public const string CcZero = "\uf348";
        public const string ConwayGlider = "\uf349";
        public const string Csharp = "\uf34a";
        public const string EmailBulk = "\uf34b";
        public const string EmailBulkO = "\uf34c";
        public const string Gnu = "\uf34d";
        public const string GooglePlay = "\uf34e";
        public const string Heroku = "\uf34f";
        public const string HomeAssistant = "\uf350";
        public const string Java = "\uf351";
        public const string Mariadb = "\uf352";
        public const string Markdown = "\uf353";
        public const string Mysql = "\uf354";
        public const string Nordcast = "\uf355";
        public const string Plume = "\uf356";
        public const string Postgresql = "\uf357";
        public const string SassAlt = "\uf359";
        public const string Sass = "\uf358";
        public const string Skate = "\uf35a";
        public const string Sketchfab = "\uf35b";
        public const string Tex = "\uf35c";
        public const string Textpattern = "\uf35d";
        public const string Unity = "\uf35e";
    }
}
