// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C#
// from https://raw.githubusercontent.com/FortAwesome/Font-Awesome/master/metadata/icons.yml
// for use with https://github.com/FortAwesome/Font-Awesome/blob/master/webfonts/fa-brands-400.ttf
namespace IconFonts
{
    public class FontAwesome5Brands
    {
        public const string FontIconFileName = "fa-brands-400.ttf";

        public const int IconMin = 0xe007;
        public const int IconMax = 0xf8e8;
        public const string The500px = "\uf26e";
        public const string AccessibleIcon = "\uf368";
        public const string Accusoft = "\uf369";
        public const string AcquisitionsIncorporated = "\uf6af";
        public const string Adn = "\uf170";
        public const string Adobe = "\uf778";
        public const string Adversal = "\uf36a";
        public const string Affiliatetheme = "\uf36b";
        public const string Airbnb = "\uf834";
        public const string Algolia = "\uf36c";
        public const string Alipay = "\uf642";
        public const string Amazon = "\uf270";
        public const string AmazonPay = "\uf42c";
        public const string Amilia = "\uf36d";
        public const string Android = "\uf17b";
        public const string Angellist = "\uf209";
        public const string Angrycreative = "\uf36e";
        public const string Angular = "\uf420";
        public const string AppStore = "\uf36f";
        public const string AppStoreIos = "\uf370";
        public const string Apper = "\uf371";
        public const string Apple = "\uf179";
        public const string ApplePay = "\uf415";
        public const string Artstation = "\uf77a";
        public const string Asymmetrik = "\uf372";
        public const string Atlassian = "\uf77b";
        public const string Audible = "\uf373";
        public const string Autoprefixer = "\uf41c";
        public const string Avianex = "\uf374";
        public const string Aviato = "\uf421";
        public const string Aws = "\uf375";
        public const string Bandcamp = "\uf2d5";
        public const string BattleNet = "\uf835";
        public const string Behance = "\uf1b4";
        public const string BehanceSquare = "\uf1b5";
        public const string Bimobject = "\uf378";
        public const string Bitbucket = "\uf171";
        public const string Bitcoin = "\uf379";
        public const string Bity = "\uf37a";
        public const string BlackTie = "\uf27e";
        public const string Blackberry = "\uf37b";
        public const string Blogger = "\uf37c";
        public const string BloggerB = "\uf37d";
        public const string Bluetooth = "\uf293";
        public const string BluetoothB = "\uf294";
        public const string Bootstrap = "\uf836";
        public const string Btc = "\uf15a";
        public const string Buffer = "\uf837";
        public const string Buromobelexperte = "\uf37f";
        public const string BuyNLarge = "\uf8a6";
        public const string Buysellads = "\uf20d";
        public const string CanadianMapleLeaf = "\uf785";
        public const string CcAmazonPay = "\uf42d";
        public const string CcAmex = "\uf1f3";
        public const string CcApplePay = "\uf416";
        public const string CcDinersClub = "\uf24c";
        public const string CcDiscover = "\uf1f2";
        public const string CcJcb = "\uf24b";
        public const string CcMastercard = "\uf1f1";
        public const string CcPaypal = "\uf1f4";
        public const string CcStripe = "\uf1f5";
        public const string CcVisa = "\uf1f0";
        public const string Centercode = "\uf380";
        public const string Centos = "\uf789";
        public const string Chrome = "\uf268";
        public const string Chromecast = "\uf838";
        public const string Cloudscale = "\uf383";
        public const string Cloudsmith = "\uf384";
        public const string Cloudversify = "\uf385";
        public const string Codepen = "\uf1cb";
        public const string Codiepie = "\uf284";
        public const string Confluence = "\uf78d";
        public const string Connectdevelop = "\uf20e";
        public const string Contao = "\uf26d";
        public const string CottonBureau = "\uf89e";
        public const string Cpanel = "\uf388";
        public const string CreativeCommons = "\uf25e";
        public const string CreativeCommonsBy = "\uf4e7";
        public const string CreativeCommonsNc = "\uf4e8";
        public const string CreativeCommonsNcEu = "\uf4e9";
        public const string CreativeCommonsNcJp = "\uf4ea";
        public const string CreativeCommonsNd = "\uf4eb";
        public const string CreativeCommonsPd = "\uf4ec";
        public const string CreativeCommonsPdAlt = "\uf4ed";
        public const string CreativeCommonsRemix = "\uf4ee";
        public const string CreativeCommonsSa = "\uf4ef";
        public const string CreativeCommonsSampling = "\uf4f0";
        public const string CreativeCommonsSamplingPlus = "\uf4f1";
        public const string CreativeCommonsShare = "\uf4f2";
        public const string CreativeCommonsZero = "\uf4f3";
        public const string CriticalRole = "\uf6c9";
        public const string Css3 = "\uf13c";
        public const string Css3Alt = "\uf38b";
        public const string Cuttlefish = "\uf38c";
        public const string DAndD = "\uf38d";
        public const string DAndDBeyond = "\uf6ca";
        public const string Dailymotion = "\ue052";
        public const string Dashcube = "\uf210";
        public const string Deezer = "\ue077";
        public const string Delicious = "\uf1a5";
        public const string Deploydog = "\uf38e";
        public const string Deskpro = "\uf38f";
        public const string Dev = "\uf6cc";
        public const string Deviantart = "\uf1bd";
        public const string Dhl = "\uf790";
        public const string Diaspora = "\uf791";
        public const string Digg = "\uf1a6";
        public const string DigitalOcean = "\uf391";
        public const string Discord = "\uf392";
        public const string Discourse = "\uf393";
        public const string Dochub = "\uf394";
        public const string Docker = "\uf395";
        public const string Draft2digital = "\uf396";
        public const string Dribbble = "\uf17d";
        public const string DribbbleSquare = "\uf397";
        public const string Dropbox = "\uf16b";
        public const string Drupal = "\uf1a9";
        public const string Dyalog = "\uf399";
        public const string Earlybirds = "\uf39a";
        public const string Ebay = "\uf4f4";
        public const string Edge = "\uf282";
        public const string EdgeLegacy = "\ue078";
        public const string Elementor = "\uf430";
        public const string Ello = "\uf5f1";
        public const string Ember = "\uf423";
        public const string Empire = "\uf1d1";
        public const string Envira = "\uf299";
        public const string Erlang = "\uf39d";
        public const string Ethereum = "\uf42e";
        public const string Etsy = "\uf2d7";
        public const string Evernote = "\uf839";
        public const string Expeditedssl = "\uf23e";
        public const string Facebook = "\uf09a";
        public const string FacebookF = "\uf39e";
        public const string FacebookMessenger = "\uf39f";
        public const string FacebookSquare = "\uf082";
        public const string FantasyFlightGames = "\uf6dc";
        public const string Fedex = "\uf797";
        public const string Fedora = "\uf798";
        public const string Figma = "\uf799";
        public const string Firefox = "\uf269";
        public const string FirefoxBrowser = "\ue007";
        public const string FirstOrder = "\uf2b0";
        public const string FirstOrderAlt = "\uf50a";
        public const string Firstdraft = "\uf3a1";
        public const string Flickr = "\uf16e";
        public const string Flipboard = "\uf44d";
        public const string Fly = "\uf417";
        public const string FontAwesome = "\uf2b4";
        public const string FontAwesomeAlt = "\uf35c";
        public const string FontAwesomeFlag = "\uf425";
        public const string FontAwesomeLogoFull = "\uf4e6";
        public const string Fonticons = "\uf280";
        public const string FonticonsFi = "\uf3a2";
        public const string FortAwesome = "\uf286";
        public const string FortAwesomeAlt = "\uf3a3";
        public const string Forumbee = "\uf211";
        public const string Foursquare = "\uf180";
        public const string FreeCodeCamp = "\uf2c5";
        public const string Freebsd = "\uf3a4";
        public const string Fulcrum = "\uf50b";
        public const string GalacticRepublic = "\uf50c";
        public const string GalacticSenate = "\uf50d";
        public const string GetPocket = "\uf265";
        public const string Gg = "\uf260";
        public const string GgCircle = "\uf261";
        public const string Git = "\uf1d3";
        public const string GitAlt = "\uf841";
        public const string GitSquare = "\uf1d2";
        public const string Github = "\uf09b";
        public const string GithubAlt = "\uf113";
        public const string GithubSquare = "\uf092";
        public const string Gitkraken = "\uf3a6";
        public const string Gitlab = "\uf296";
        public const string Gitter = "\uf426";
        public const string Glide = "\uf2a5";
        public const string GlideG = "\uf2a6";
        public const string Gofore = "\uf3a7";
        public const string Goodreads = "\uf3a8";
        public const string GoodreadsG = "\uf3a9";
        public const string Google = "\uf1a0";
        public const string GoogleDrive = "\uf3aa";
        public const string GooglePay = "\ue079";
        public const string GooglePlay = "\uf3ab";
        public const string GooglePlus = "\uf2b3";
        public const string GooglePlusG = "\uf0d5";
        public const string GooglePlusSquare = "\uf0d4";
        public const string GoogleWallet = "\uf1ee";
        public const string Gratipay = "\uf184";
        public const string Grav = "\uf2d6";
        public const string Gripfire = "\uf3ac";
        public const string Grunt = "\uf3ad";
        public const string Gulp = "\uf3ae";
        public const string HackerNews = "\uf1d4";
        public const string HackerNewsSquare = "\uf3af";
        public const string Hackerrank = "\uf5f7";
        public const string Hips = "\uf452";
        public const string HireAHelper = "\uf3b0";
        public const string Hooli = "\uf427";
        public const string Hornbill = "\uf592";
        public const string Hotjar = "\uf3b1";
        public const string Houzz = "\uf27c";
        public const string Html5 = "\uf13b";
        public const string Hubspot = "\uf3b2";
        public const string Ideal = "\ue013";
        public const string Imdb = "\uf2d8";
        public const string Instagram = "\uf16d";
        public const string InstagramSquare = "\ue055";
        public const string Intercom = "\uf7af";
        public const string InternetExplorer = "\uf26b";
        public const string Invision = "\uf7b0";
        public const string Ioxhost = "\uf208";
        public const string ItchIo = "\uf83a";
        public const string Itunes = "\uf3b4";
        public const string ItunesNote = "\uf3b5";
        public const string Java = "\uf4e4";
        public const string JediOrder = "\uf50e";
        public const string Jenkins = "\uf3b6";
        public const string Jira = "\uf7b1";
        public const string Joget = "\uf3b7";
        public const string Joomla = "\uf1aa";
        public const string Js = "\uf3b8";
        public const string JsSquare = "\uf3b9";
        public const string Jsfiddle = "\uf1cc";
        public const string Kaggle = "\uf5fa";
        public const string Keybase = "\uf4f5";
        public const string Keycdn = "\uf3ba";
        public const string Kickstarter = "\uf3bb";
        public const string KickstarterK = "\uf3bc";
        public const string Korvue = "\uf42f";
        public const string Laravel = "\uf3bd";
        public const string Lastfm = "\uf202";
        public const string LastfmSquare = "\uf203";
        public const string Leanpub = "\uf212";
        public const string Less = "\uf41d";
        public const string Line = "\uf3c0";
        public const string Linkedin = "\uf08c";
        public const string LinkedinIn = "\uf0e1";
        public const string Linode = "\uf2b8";
        public const string Linux = "\uf17c";
        public const string Lyft = "\uf3c3";
        public const string Magento = "\uf3c4";
        public const string Mailchimp = "\uf59e";
        public const string Mandalorian = "\uf50f";
        public const string Markdown = "\uf60f";
        public const string Mastodon = "\uf4f6";
        public const string Maxcdn = "\uf136";
        public const string Mdb = "\uf8ca";
        public const string Medapps = "\uf3c6";
        public const string Medium = "\uf23a";
        public const string MediumM = "\uf3c7";
        public const string Medrt = "\uf3c8";
        public const string Meetup = "\uf2e0";
        public const string Megaport = "\uf5a3";
        public const string Mendeley = "\uf7b3";
        public const string Microblog = "\ue01a";
        public const string Microsoft = "\uf3ca";
        public const string Mix = "\uf3cb";
        public const string Mixcloud = "\uf289";
        public const string Mixer = "\ue056";
        public const string Mizuni = "\uf3cc";
        public const string Modx = "\uf285";
        public const string Monero = "\uf3d0";
        public const string Napster = "\uf3d2";
        public const string Neos = "\uf612";
        public const string Nimblr = "\uf5a8";
        public const string Node = "\uf419";
        public const string NodeJs = "\uf3d3";
        public const string Npm = "\uf3d4";
        public const string Ns8 = "\uf3d5";
        public const string Nutritionix = "\uf3d6";
        public const string Odnoklassniki = "\uf263";
        public const string OdnoklassnikiSquare = "\uf264";
        public const string OldRepublic = "\uf510";
        public const string Opencart = "\uf23d";
        public const string Openid = "\uf19b";
        public const string Opera = "\uf26a";
        public const string OptinMonster = "\uf23c";
        public const string Orcid = "\uf8d2";
        public const string Osi = "\uf41a";
        public const string Page4 = "\uf3d7";
        public const string Pagelines = "\uf18c";
        public const string Palfed = "\uf3d8";
        public const string Patreon = "\uf3d9";
        public const string Paypal = "\uf1ed";
        public const string PennyArcade = "\uf704";
        public const string Periscope = "\uf3da";
        public const string Phabricator = "\uf3db";
        public const string PhoenixFramework = "\uf3dc";
        public const string PhoenixSquadron = "\uf511";
        public const string Php = "\uf457";
        public const string PiedPiper = "\uf2ae";
        public const string PiedPiperAlt = "\uf1a8";
        public const string PiedPiperHat = "\uf4e5";
        public const string PiedPiperPp = "\uf1a7";
        public const string PiedPiperSquare = "\ue01e";
        public const string Pinterest = "\uf0d2";
        public const string PinterestP = "\uf231";
        public const string PinterestSquare = "\uf0d3";
        public const string Playstation = "\uf3df";
        public const string ProductHunt = "\uf288";
        public const string Pushed = "\uf3e1";
        public const string Python = "\uf3e2";
        public const string Qq = "\uf1d6";
        public const string Quinscape = "\uf459";
        public const string Quora = "\uf2c4";
        public const string RProject = "\uf4f7";
        public const string RaspberryPi = "\uf7bb";
        public const string Ravelry = "\uf2d9";
        public const string React = "\uf41b";
        public const string Reacteurope = "\uf75d";
        public const string Readme = "\uf4d5";
        public const string Rebel = "\uf1d0";
        public const string RedRiver = "\uf3e3";
        public const string Reddit = "\uf1a1";
        public const string RedditAlien = "\uf281";
        public const string RedditSquare = "\uf1a2";
        public const string Redhat = "\uf7bc";
        public const string Renren = "\uf18b";
        public const string Replyd = "\uf3e6";
        public const string Researchgate = "\uf4f8";
        public const string Resolving = "\uf3e7";
        public const string Rev = "\uf5b2";
        public const string Rocketchat = "\uf3e8";
        public const string Rockrms = "\uf3e9";
        public const string Rust = "\ue07a";
        public const string Safari = "\uf267";
        public const string Salesforce = "\uf83b";
        public const string Sass = "\uf41e";
        public const string Schlix = "\uf3ea";
        public const string Scribd = "\uf28a";
        public const string Searchengin = "\uf3eb";
        public const string Sellcast = "\uf2da";
        public const string Sellsy = "\uf213";
        public const string Servicestack = "\uf3ec";
        public const string Shirtsinbulk = "\uf214";
        public const string Shopify = "\ue057";
        public const string Shopware = "\uf5b5";
        public const string Simplybuilt = "\uf215";
        public const string Sistrix = "\uf3ee";
        public const string Sith = "\uf512";
        public const string Sketch = "\uf7c6";
        public const string Skyatlas = "\uf216";
        public const string Skype = "\uf17e";
        public const string Slack = "\uf198";
        public const string SlackHash = "\uf3ef";
        public const string Slideshare = "\uf1e7";
        public const string Snapchat = "\uf2ab";
        public const string SnapchatGhost = "\uf2ac";
        public const string SnapchatSquare = "\uf2ad";
        public const string Soundcloud = "\uf1be";
        public const string Sourcetree = "\uf7d3";
        public const string Speakap = "\uf3f3";
        public const string SpeakerDeck = "\uf83c";
        public const string Spotify = "\uf1bc";
        public const string Squarespace = "\uf5be";
        public const string StackExchange = "\uf18d";
        public const string StackOverflow = "\uf16c";
        public const string Stackpath = "\uf842";
        public const string Staylinked = "\uf3f5";
        public const string Steam = "\uf1b6";
        public const string SteamSquare = "\uf1b7";
        public const string SteamSymbol = "\uf3f6";
        public const string StickerMule = "\uf3f7";
        public const string Strava = "\uf428";
        public const string Stripe = "\uf429";
        public const string StripeS = "\uf42a";
        public const string Studiovinari = "\uf3f8";
        public const string Stumbleupon = "\uf1a4";
        public const string StumbleuponCircle = "\uf1a3";
        public const string Superpowers = "\uf2dd";
        public const string Supple = "\uf3f9";
        public const string Suse = "\uf7d6";
        public const string Swift = "\uf8e1";
        public const string Symfony = "\uf83d";
        public const string Teamspeak = "\uf4f9";
        public const string Telegram = "\uf2c6";
        public const string TelegramPlane = "\uf3fe";
        public const string TencentWeibo = "\uf1d5";
        public const string TheRedYeti = "\uf69d";
        public const string Themeco = "\uf5c6";
        public const string Themeisle = "\uf2b2";
        public const string ThinkPeaks = "\uf731";
        public const string Tiktok = "\ue07b";
        public const string TradeFederation = "\uf513";
        public const string Trello = "\uf181";
        public const string Tripadvisor = "\uf262";
        public const string Tumblr = "\uf173";
        public const string TumblrSquare = "\uf174";
        public const string Twitch = "\uf1e8";
        public const string Twitter = "\uf099";
        public const string TwitterSquare = "\uf081";
        public const string Typo3 = "\uf42b";
        public const string Uber = "\uf402";
        public const string Ubuntu = "\uf7df";
        public const string Uikit = "\uf403";
        public const string Umbraco = "\uf8e8";
        public const string Uniregistry = "\uf404";
        public const string Unity = "\ue049";
        public const string Unsplash = "\ue07c";
        public const string Untappd = "\uf405";
        public const string Ups = "\uf7e0";
        public const string Usb = "\uf287";
        public const string Usps = "\uf7e1";
        public const string Ussunnah = "\uf407";
        public const string Vaadin = "\uf408";
        public const string Viacoin = "\uf237";
        public const string Viadeo = "\uf2a9";
        public const string ViadeoSquare = "\uf2aa";
        public const string Viber = "\uf409";
        public const string Vimeo = "\uf40a";
        public const string VimeoSquare = "\uf194";
        public const string VimeoV = "\uf27d";
        public const string Vine = "\uf1ca";
        public const string Vk = "\uf189";
        public const string Vnv = "\uf40b";
        public const string Vuejs = "\uf41f";
        public const string Waze = "\uf83f";
        public const string Weebly = "\uf5cc";
        public const string Weibo = "\uf18a";
        public const string Weixin = "\uf1d7";
        public const string Whatsapp = "\uf232";
        public const string WhatsappSquare = "\uf40c";
        public const string Whmcs = "\uf40d";
        public const string WikipediaW = "\uf266";
        public const string Windows = "\uf17a";
        public const string Wix = "\uf5cf";
        public const string WizardsOfTheCoast = "\uf730";
        public const string WolfPackBattalion = "\uf514";
        public const string Wordpress = "\uf19a";
        public const string WordpressSimple = "\uf411";
        public const string Wpbeginner = "\uf297";
        public const string Wpexplorer = "\uf2de";
        public const string Wpforms = "\uf298";
        public const string Wpressr = "\uf3e4";
        public const string Xbox = "\uf412";
        public const string Xing = "\uf168";
        public const string XingSquare = "\uf169";
        public const string YCombinator = "\uf23b";
        public const string Yahoo = "\uf19e";
        public const string Yammer = "\uf840";
        public const string Yandex = "\uf413";
        public const string YandexInternational = "\uf414";
        public const string Yarn = "\uf7e3";
        public const string Yelp = "\uf1e9";
        public const string Yoast = "\uf2b1";
        public const string Youtube = "\uf167";
        public const string YoutubeSquare = "\uf431";
        public const string Zhihu = "\uf63f";
    }
}
